import itertools

import numpy as np
from collections import namedtuple
import McUtils.Devutils as dev
import McUtils.Coordinerds as coordops
import McUtils.Numputils as nput

__all__ = [
    "StateMaker",
    "modify_hamiltonian",
    "modify_internal_hamiltonian"
]

class StateMaker:
    """
    A tiny but useful class to make states based on their quanta
    of excitation
    """

    def __init__(self, ndim, mode='low-high'):
        self.ndim = ndim
        self.mode = mode

    @classmethod
    def parse_state(cls, state, mode='low-high'):
        nzp = np.nonzero(state)
        if len(nzp) > 0: nzp = nzp[0]
        if len(nzp) == 0: return "()"
        if mode == 'low-high':
            pos = len(state) - nzp
            state_quants = reversed(list(zip(pos, nzp)))
        elif mode == 'high-low':
            pos = nzp + 1
            state_quants = list(zip(pos, nzp))
        else:
            pos = nzp
            state_quants = list(zip(pos, nzp))
        return "".join("{}({})".format(p, state[i]) for p,i in state_quants)

    def make_state(self, *specs, mode=None):

        if mode is None:
            mode = self.mode

        state = [0] * self.ndim
        for s in specs:
            if isinstance(s, (int, np.integer)):
                i = s
                q = 1
            else:
                i,q = s
            if mode == 'low-high':
                state[-i] = q
            elif mode == 'high-low':
                state[i-1] = q
            elif mode == 'normal':
                state[i] = q
            else:
                raise ValueError("don't know what to do with filling mode '{}'".format(mode))
        return state

    def __call__(self, *specs, mode=None):
        return self.make_state(*specs, mode=mode)

def _get_tag_match(tag, mod_types):
    return mod_types.get(tag)

def modify_hamiltonian(ham,
                       index_specifiers,
                       index_key_function=None,
                       coupling_key_function=None,
                       scaling_types=None,
                       coupling_types=None,
                       shift_types=None,
                       tag_match=None
                       ):

    ham = np.array(ham)
    if scaling_types is None:
        scaling_types = {}
    if coupling_types is None:
        coupling_types = {}
    if shift_types is None:
        shift_types = {}
    if index_key_function is None:
        index_key_function = lambda i,t:(t1, (t1,))
    if coupling_key_function is None:
        coupling_key_function = lambda ij,t1,t2:((t1,t2), (t2, t1))

    if tag_match is None:
        tag_match = _get_tag_match

    for i,t1 in enumerate(index_specifiers):
        keys = index_key_function(i,t1)
        for t in keys:
            s = tag_match(t, coupling_types)
            if s is not None:
                ham[i, i] = s
                # break
            s = tag_match(t, scaling_types)
            if s is not None:
                ham[i, i] *= s
                # break
            s = tag_match(t, shift_types)
            if s is not None:
                ham[i, i] += s
                # break

        for j,t2 in enumerate(index_specifiers[i+1:]):
            k = j + i + 1
            keys = coupling_key_function((i,k), t1, t2)
            for t in keys:
                s = tag_match(t, coupling_types)
                if s is not None:
                    ham[i, k] = s
                    ham[k, i] = s
                    # break
                s = tag_match(t, scaling_types)
                if s is not None:
                    ham[i, k] *= s
                    ham[k, i] *= s
                    # break
                s = tag_match(t, shift_types)
                if s is not None:
                    ham[i, k] += s
                    ham[k, i] += s
                    # break

    return ham


product_state = namedtuple("product_state", ["indices", "quanta", "tags"])
def _canonicalize_internal_state_key(k):
    if isinstance(k, str):
        return product_state(None, None, (k,))
    elif all(isinstance(s, str) for s in k):
        return product_state(None, None, (tuple(k),))
    elif all(nput.is_numeric(x) for x in k):
        return product_state(
            tuple(n for n,s in enumerate(k) if s != 0),
            tuple(s for n,s in enumerate(k) if s != 0),
            None
        )
    else:
        if nput.is_numeric(k[0]):
            k = [k]
        return product_state(
            None,
            tuple(n for n,s in k),
            tuple(s for n,s in k)
        )

def _tag_match(tags1, tags2):
    if tags2 is None or tags1 is None: return False
    ml = min([len(tags1), len(tags2)])
    return tags1[-ml:] == tags2[-ml:]


def _product_state_match(state1:product_state, state2:product_state):
    if (
            state1.indices is not None
            and state2.indices is not None
    ):
        if state1.indices != state2.indices: return False
        if state1.quanta is None: return True
        if state2.quanta is None: return True
        return state1.quanta == state2.quanta
    else:
        if (
                state2.tags is None
                or state1.tags is None
        ):
            return (
                    state1.quanta is None
                    or state2.quanta is None
                    or list(sorted(state1.quanta)) == list(sorted(state2.quanta))
            )
        else:
            if len(state1.tags) != len(state2.tags): return False
            if state1.quanta is None or state2.quanta is None:
                # any match is valid of these can match
                for p in itertools.permutations(range(len(state1.tags)), len(state1.tags)):
                    if all(
                        _tag_match(s1, s2)
                        for s1, s2 in zip(
                            [state1.tags[i] for i in p],
                            state2.tags
                        )
                    ):
                        return p
            else:
                for p in itertools.permutations(range(len(state1.tags)), len(state1.tags)):
                    if all(
                            (
                                    n1 is None or n2 is None
                                    or n1 == n2
                            ) and _tag_match(s1, s2)
                            for n1,s1, n2,s2 in zip(
                                [state1.quanta[i] for i in p],
                                [state1.tags[i] for i in p],
                                state2.quanta,
                                state2.tags
                            )
                    ):
                        return p

def _canonicalize_scaling_key(key):
    if isinstance(key, product_state):
        return key
    elif (
            len(key) == 3
            and isinstance(key[1], product_state)
            and isinstance(key[2], product_state)
    ):
        shats = key[0]
        if nput.is_numeric(shats):
            shats = [[shats]]
        else:
            shats = [[s] if nput.is_numeric(s) else s for s in shats]
        shats = tuple(tuple(int(i) for i in s) for s in shats)
        return (
            shats,
            key[1],
            key[2]
        )
    elif (
            len(key) == 2
            and isinstance(key[0], product_state)
            and isinstance(key[1], product_state)
    ):
        return (
            key[0],
            key[1]
        )
    if (
            isinstance(key, str)
            or all(isinstance(k, str) for k in key)
            or all(nput.is_numeric(x) for x in key)
    ):
        return _canonicalize_internal_state_key(key)
    if len(key) == 3:
        shats = key[0]
        if nput.is_numeric(shats):
            shats = [[shats]]
        else:
            shats = [[s] if nput.is_numeric(s) else s for s in shats]
        shats = tuple(tuple(int(i) for i in s) for s in shats)
        return (
            shats,
            _canonicalize_internal_state_key(key[1]),
            _canonicalize_internal_state_key(key[2])
        )
    else:
        return (
            _canonicalize_internal_state_key(key[0]),
            _canonicalize_internal_state_key(key[1])
        )

def _find_internal_coupling_match(spec, couplings):
    base = couplings.get(spec)
    if base is not None: return base
    if nput.is_numeric(spec) or not (
            isinstance(spec, product_state)
            or any(isinstance(s, product_state) for s in spec)
    ): return None


    if isinstance(spec, product_state):
        for s,v in couplings.items():
            if isinstance(s, product_state) and _product_state_match(s, spec):
                return v
    else:
        if not isinstance(spec[0], product_state):
            shared_ats = spec[0]
            spec = spec[1:]
        else:
            shared_ats = None

        spec1, spec2 = spec
        for s,v in couplings.items():
            if not isinstance(s, product_state):
                if not isinstance(s[0], product_state):
                    shared_ats_test = s[0]
                    s = s[1:]
                else:
                    shared_ats_test = None

                good_test = (
                    shared_ats_test is None and
                    shared_ats is None
                ) or (
                        shared_ats_test is not None and
                        shared_ats is not None and
                        list(sorted(len(t) for t in shared_ats_test))
                            == list(sorted(len(t) for t in shared_ats))
                )
                if not good_test: continue

                s1,s2 = s
                perm_1 = _product_state_match(s1, spec1)
                perm_2 = _product_state_match(s2, spec2)
                if perm_1 is None or perm_2 is None:
                    perm_1 = _product_state_match(s1, spec2)
                    perm_2 = _product_state_match(s2, spec1)

                if not perm_1: continue
                if not perm_2: continue

                if perm_1 is not None and perm_2 is not None:
                    if shared_ats is not None:
                        if perm_1 is True or perm_2 is True:
                            return list(sorted(shared_ats)) == list(sorted(shared_ats_test))
                        elif all(
                                shared_ats[p1] == shared_ats_test[p2]
                                for p1,p2 in zip(perm_1, perm_2)
                        ):
                            return v
                    else:
                        return v

def _internal_index_key_function(internals, tags, atoms):
    def index_key_function(i, state):
        return (i,
                (state,),
                (state, state),
                product_state(
                    tuple(n for n, s in enumerate(state) if s != 0),
                    tuple(s for n, s in enumerate(state) if s != 0),
                    tuple(tags[n] for n, s in enumerate(state) if s != 0)
                )
                )
    return index_key_function

def _internal_coupling_key_function(internals, tags, atoms):
    def coupling_key_function(ij, state1, state2):
        i, j = ij
        matches = []
        for n1,s1 in enumerate(state1):
            if s1 == 0: continue
            sublist = []
            for n2,s2 in enumerate(state2):
                if s2 == 0: continue
                internal_1 = internals[n1]
                internal_2 = internals[n2]
                if not (nput.is_atomic(internal_1) or  nput.is_atomic(internal_2)):
                    num_shared = sum(
                        1 if i1 in internal_2 else 0
                        for i1 in internal_1
                    )
                else:
                    num_shared = 0
                sublist.append(num_shared)
            matches.append(tuple(sublist))

        return (
            (i, j),
            (state1, state2),
            (state2, state1),

            (
                tuple(matches),
                product_state(
                    tuple(n for n, s in enumerate(state1) if s != 0),
                    tuple(s for n, s in enumerate(state1) if s != 0),
                    tuple(tags[n] for n, s in enumerate(state1) if s != 0)
                ),
                product_state(
                    tuple(n for n, s in enumerate(state2) if s != 0),
                    tuple(s for n, s in enumerate(state2) if s != 0),
                    tuple(tags[n] for n, s in enumerate(state2) if s != 0)
                )
            ),
            (
                product_state(
                    tuple(n for n, s in enumerate(state1) if s != 0),
                    tuple(s for n, s in enumerate(state1) if s != 0),
                    tuple(tags[n] for n, s in enumerate(state1) if s != 0)
                ),
                product_state(
                    tuple(n for n, s in enumerate(state2) if s != 0),
                    tuple(s for n, s in enumerate(state2) if s != 0),
                    tuple(tags[n] for n, s in enumerate(state2) if s != 0)
                )
            )

        )
    return coupling_key_function


def modify_internal_hamiltonian(ham,
                                internals,
                                states=None,
                                atoms=None,
                                tags=None,
                                index_key_function=None,
                                coupling_key_function=None,
                                scaling_types=None,
                                coupling_types=None,
                                shift_types=None,
                                tag_match=None
                                ):
    if hasattr(internals, 'items'):
        internals = {
            (
                coordops.canonicalize_internal(k)
                    if not dev.is_atomic(k) and all(nput.is_int(cc) for cc in k) else
                k
            ):t
            for k,t in internals.items()
        }
        if tags is None:
            tags = list(internals.values())
        internals = list(internals.keys())
    else:
        if tags is None:
            tags = [
                list(i.values())[0]
                    if hasattr(i, 'values') else
                None
                for i in internals
            ]
        internals = [
            i
                if hasattr(i, 'values') or dev.is_atomic(i) or not all(nput.is_int(cc) for cc in i) else
            coordops.canonicalize_internal(i)
            for i in internals
        ]

    if states is None:
        states = np.eye(ham.shape[0], dtype=int)

    if index_key_function is None:
        index_key_function = _internal_index_key_function(internals, tags, atoms)
    if coupling_key_function is None:
        coupling_key_function = _internal_coupling_key_function(internals, tags, atoms)

    if scaling_types is not None:
        scaling_types = {
            _canonicalize_scaling_key(k):v
            for k,v in scaling_types.items()
        }
    if coupling_types is not None:
        coupling_types = {
            _canonicalize_scaling_key(k):v
            for k,v in coupling_types.items()
        }
    if shift_types is not None:
        shift_types = {
            _canonicalize_scaling_key(k):v
            for k,v in shift_types.items()
        }

    if tag_match is None:
        tag_match = _find_internal_coupling_match

    return modify_hamiltonian(ham,
                              [tuple(np.asanyarray(s).astype(int)) for s in states],
                              index_key_function=index_key_function,
                              coupling_key_function=coupling_key_function,
                              scaling_types=scaling_types,
                              coupling_types=coupling_types,
                              shift_types=shift_types,
                              tag_match=tag_match
                              )



"""
Provides a general class of localized modes using a potentially non-linear
transformation of normal modes
"""

import numpy as np, scipy
import McUtils.Numputils as nput

from .MixtureModes import *
from .NormalModes import *

__all__ = ['LocalizedModes']
class LocalizedModes(MixtureModes):

    def __init__(self,
                 normal_modes: MixtureModes,
                 transformation,
                 inverse=None,
                 origin=None,
                 masses=None,
                 freqs=None,
                 mass_weighted=None,
                 frequency_scaled=None,
                 **etc
                 ):
        if len(transformation) == 2 and nput.is_array_like(transformation[0]):
            tf, inv = transformation
            tf = np.asanyarray(tf)
            if tf.ndim == 2:
                transformation = tf
                if inverse is None:
                    inverse = np.asanyarray(inv)
        else:
            transformation = np.asanyarray(transformation)
            if inverse is None:
                inverse = transformation.T
        mat = normal_modes.modes_by_coords @ transformation
        self.localizing_transformation = (transformation, inverse)
        self.base_modes = normal_modes
        inverse = inverse @ normal_modes.coords_by_modes
        if origin is None:
            origin = normal_modes.origin
        if masses is None:
            masses = normal_modes.masses
        super().__init__(
            normal_modes.basis,
            mat,
            inverse=inverse,
            origin=origin,
            masses=masses,
            freqs=normal_modes.freqs,
            mass_weighted=normal_modes.mass_weighted,
            frequency_scaled=frequency_scaled,
            **etc
        )
        self._freqs = freqs

    def __getitem__(self, item):
        """
        Takes a slice of the modes
        :param item:
        :type item:
        :return:
        :rtype:
        """

        if isinstance(item, (int, np.integer)):
            item = (item,)
        elif isinstance(item, slice):
            ...
        elif not isinstance(item[0], (int, np.integer)):
            item = tuple(item[0])

        # sub_modes = self.matrix[:, item]
        # inv = self._inv
        # if inv is not None:
        #     inv = inv[item, :]
        # freq = self.freqs[item,]
        tf, inv = self.localizing_transformation
        hmm = self.modify(
            transformation=(tf[:, item], inv[item, :])
        )

        return hmm

    @property
    def freqs(self):
        if self._freqs is None:
            self._freqs = self.compute_freqs()
        return self._freqs
    @freqs.setter
    def freqs(self, freqs):
        ...
        # if self._in_init:
        #     ...
        # else:
        #     self._freqs = freqs
    @property
    def mass_weighted(self):
        return self.base_modes.mass_weighted
    @mass_weighted.setter
    def mass_weighted(self, new):
        if (
                new and not self.base_modes.mass_weighted
                or not new and self.base_modes.mass_weighted
        ):
            raise ValueError("can't set `mass_weighted` directly")
    # @property
    # def frequency_scaled(self):
    #     return self.base_modes.frequency_scaled
    # @frequency_scaled.setter
    # def frequency_scaled(self, new):
    #     if (
    #             new and not self.base_modes.frequency_scaled
    #             or not new and self.base_modes.frequency_scaled
    #     ):
    #         raise ValueError("can't set `frequency_scaled` directly")
    @property
    def g_matrix(self):
        return self.base_modes.g_matrix
    #     if self.base_modes.g_matrix is not None:
    #         tf, inv = self.localizing_transformation
    #         base_gm = self.base_modes.g_matrix
    #         if base_gm is None:
    #             base_gm = self.base_modes.compute_gmatrix()
    #         pinv = self.modes_by_coords @ inv.T
    #         return self.coords_by_modes @ inv.T @ base_gm @ inv
    #     else:
    #         return None
    @g_matrix.setter
    def g_matrix(self, g):
        ...

    def modify(self,
               base_modes=None,
               *,
               transformation=None,
               freqs=None,
               origin=None,
               masses=None,
               inverse=None,
               name=None,
               mass_weighted=None,
               frequency_scaled=None,
               g_matrix=None
               ):
        if (
                transformation is not None
                and inverse is None
                and len(transformation) == 2
                and nput.is_array_like(transformation[0])
        ):
            tf, inv = transformation
            tf = np.asanyarray(tf)
            if tf.ndim == 2:
                transformation = tf
                inverse = inv
        return type(self)(
            self.base_modes if base_modes is None else base_modes,
            self.localizing_transformation[0] if transformation is None else transformation,
            inverse=self.localizing_transformation[1] if inverse is None else inverse,
            origin=origin,
            masses=masses,
            freqs=freqs,
            g_matrix=g_matrix,
            name=self.name if name is None else name,
            frequency_scaled=frequency_scaled
        )

    def _frequency_scaling(self, freqs=None):
        # L = self.matrix.shape.T
        if freqs is None:
            freqs = self.local_freqs
        conv = np.sqrt(freqs)
        return freqs, conv

    def make_mass_weighted(self, **kwargs):
        return self.modify(self.base_modes.make_mass_weighted(**kwargs))
    def remove_mass_weighting(self, **kwargs):
        return self.modify(self.base_modes.remove_mass_weighting(**kwargs))
    def make_frequency_scaled(self, freqs=None, **kwargs):
        if self.frequency_scaled: return self
        freqs, conv = self._frequency_scaling(freqs=freqs)
        tf, inv = self.localizing_transformation

        return self.modify(
            transformation=(tf@np.diag(conv), np.diag(1/conv)@inv),
            frequency_scaled=True
        )
    def remove_frequency_scaling(self, freqs=None, **kwargs):
        if not self.frequency_scaled: return self
        freqs, conv = self._frequency_scaling(freqs=freqs)
        tf, inv = self.localizing_transformation

        return self.modify(
            transformation=(tf@np.diag(1/conv), np.diag(conv)@inv),
            frequency_scaled=False
        )

    # @property
    # def local_hessian(self):
    #     tf, inv = self.localizing_transformation
    #     f = inv @ np.diag(self.freqs ** 2) @ inv.T
    #     g = self.g_matrix
    #     a = np.diag(np.power(np.diag(g) / np.diag(f), 1 / 4))
    #     return a @ f @ a

    def compute_hessian(self, system='modes'):
        if system == 'modes':
            tf, inv = self.localizing_transformation
            return inv @ np.diag(np.sign(self.base_modes.freqs) * (self.base_modes.freqs ** 2)) @ inv.T
        elif system == 'coords':
            return self.base_modes.compute_hessian('coords')
        else:
            raise ValueError(f'unknown system for normal modes "{system}", valid are "modes", "coords"')

    # def localize(self,
    #              method=None,
    #              **opts
    #              ):
    #     return NormalModes.localize(self, method=method, **opts)

    def apply_transformation(self, transformation, inverse=None, **opts):
        if len(transformation) == 2 and nput.is_array_like(transformation[0]):
            tf, inv = transformation
            tf = np.asanyarray(tf)
            if tf.ndim == 2:
                transformation = tf
                if inverse is None:
                    inverse = np.asanyarray(inv)
        else:
            transformation = np.asanyarray(transformation)
            if inverse is None:
                inverse = transformation.T
        tf_1, inv_1 = self.localizing_transformation
        transformation = tf_1 @ transformation
        inverse = inverse @ inv_1

        return type(self)(
            self.base_modes,
            transformation,
            inverse=inverse,
            **opts
        )

    def get_complement(self, concatenate=False):
        rem = nput.find_basis(
            nput.orthogonal_projection_matrix(self.localizing_transformation[0])
        )
        submodes = self.base_modes.apply_transformation(rem)
        if not concatenate:
            loc_freqs, loc_modes = scipy.linalg.eigh(submodes.compute_hessian(), submodes.compute_gmatrix(), type=3)
            return submodes.apply_transformation(loc_modes)
        else:
            g = submodes.compute_gmatrix()
            g12 = nput.fractional_power(g, 1/2)
            gi12 = nput.fractional_power(g, -1/2)
            f = submodes.compute_hessian()
            _, mw_modes = np.linalg.eigh(g12 @ f @ g12)
            modes, inv_loc = g12 @ mw_modes, mw_modes.T @ gi12
            # loc_freqs, loc_modes = scipy.linalg.eigh(submodes.compute_hessian(), submodes.compute_gmatrix(), type=3)
            tf_rem = rem @ modes
            tf, inv = self.localizing_transformation
            inv_rem = inv_loc @ rem.T

            tf_full = np.concatenate([tf, tf_rem], axis=1)
            inv_full = np.concatenate([inv, inv_rem], axis=0)

            return self.base_modes.apply_transformation((tf_full, inv_full))
        # return self.base_modes.localize(
        #     projections=[self.make_mass_weighted().modes_by_coords],
        #     orthogonal_projection=True
        # )


import numpy as np

import McUtils.Numputils as nput
import McUtils.Combinatorics as comb

def nested_commutator_expansion(k, side='left'):
    if side == 'left':
        comm = 0
        for i in range(1, k):
            comm = [comm, i]
    else:
        comm = k - 1
        for i in range(k - 2, -1, -1):
            comm = [i, comm]
    return comm

def closed_system_propagator(interaction_generator, time_delays):
    iteractions = [interaction_generator(t) for t in time_delays]
    interaction_n = len(iteractions)
    comms = nput.commutator_terms(nested_commutator_expansion(interaction_n))
    #TODO: allow for pathway filtering on the commutator terms
    return nput.commutator_evaluate(
        comms, iteractions,
        direct=False,
        normalized=True
    )

def liouville_pathways(k):
    final_states = [[i, k-i] for i in range(k+1)]
    return np.concatenate([ # just lattice paths, but people don't call them that
        comb.UniquePermutations([0]*l + [1]*r)
        for l,r in final_states
    ])

def evaluate_pathway(pathway, initial_state, transition_graph):
    ...

def liouville_pathway_propagator(pathways):
    signs = (-1)**np.sum(pathways, axis=1)
    signal = ...
    for s,p in pathways:
        ...

def kronecker_delta_product_evaluate():
    ...

def isotropic_tensor_basis(k):
    ...

def rotational_averaging_data(k) -> (dict, np.ndarray, dict):
    ...

def rotational_average(isotropic_basis):
    ...



__all__ = [
    "MolecularHamiltonian"
]

import numpy as np, itertools
import McUtils.Numputils as nput
import McUtils.Iterators as itut
from McUtils.Combinatorics import UniquePermutations
from McUtils.Zachary import TensorDerivativeConverter

from ..Modes import MixtureModes

from .CoordinateSystems import MolecularEmbedding, ModeEmbedding
from .Properties import PotentialSurfaceManager, DipoleSurfaceManager, NormalModesManager

class MolecularHamiltonian:
    """
    Provides a generic molecular Hamiltonian with support for expansions (for VPT),
    construction of molecular models, and other wave function generators
    """

    def __init__(self,
                 embedding:MolecularEmbedding,
                 potential_manager:PotentialSurfaceManager=None,
                 potential_derivatives:list[np.ndarray]=None,
                 modes_manager:NormalModesManager=None,
                 modes:MixtureModes=None,
                 dipole_manager:DipoleSurfaceManager=None,
                 dipole_derivatives:list[np.ndarray]=None,
                 ):
        self.embedding = embedding
        self.modes = self._prep_modes(modes_manager, modes)
        self.potential = self._prep_potential(potential_manager, potential_derivatives)
        self.dipole = self._prep_dipole(dipole_manager, dipole_derivatives)

    @classmethod
    def _prep_modes(cls, modes_manager, modes):
        if modes_manager is not None:
            return modes_manager
        else:
            return modes

    @classmethod
    def _prep_potential(cls, potential_manager, potential_derivatives):
        if potential_manager is not None:
            return ScalarOperatorManager(potential_manager)
        else:
            return ScalarOperatorManager(potential_derivatives)

    @classmethod
    def _prep_dipole(cls, dipole_manager, dipole_derivatives):
        if dipole_manager is not None:
            return ScalarOperatorManager(dipole_manager)
        else:
            return ScalarOperatorManager(dipole_derivatives)

    def get_VPT_expansions(self,
                           order=2,
                           coordinate_transformation=None
                           ):
        carts = self.embedding.internals is None
        embedding = ModeEmbedding(self.embedding, self.modes, dimensionless=True)
        if isinstance(order, (int, np.integer)):
            n = order
            order = {
                'potential':n,
                'kinetic':n,
            }
            if carts:
                order['coriolis'] = n-2
                order['watson'] = n-2
            else:
                order['pseudopotential'] = n-2

        coordinate_transformation = self.prep_transformation(coordinate_transformation)

        expansions = {}
        for k,o in order.items():
            post_processor = None
            if k == 'kinetic':
                generator = self.gmatrix_expansion(embedding=embedding)
            elif k == 'potential':
                # ders = self.potential.derivatives
                # if np.sum(np.abs(ders[0])) > 1e-8: ...
                generator = self.potential_expansion(embedding=embedding, exclude_gradient=True)
                post_processor = lambda exp: exp[1:]
            elif k == 'coriolis':
                generator = self.coriolis_expansion(embedding=embedding)
            elif k == 'pseudopotential':
                generator = self.pseudopotential_expansion(embedding)
            elif k == 'watson':
                generator = self.watson_expansion(embedding)
            else:
                raise ValueError("don't know what to do with key {}".format(k))
            terms = generator.get_terms(o, transformation=coordinate_transformation)
            if post_processor is not None:
                terms = post_processor(terms)
            expansions[k] = terms

        return expansions

    @classmethod
    def prep_transformation(cls, coordinate_transformation):
        if coordinate_transformation is not None:
            if nput.is_numeric(coordinate_transformation[0][0][0]):
                coordinate_transformation = (
                    coordinate_transformation,
                    nput.inverse_transformation(coordinate_transformation, len(coordinate_transformation))
                )

        return coordinate_transformation

    def get_embedding(self, masses=None, modes=True, mass_weight=True, dimensionless=False):
        if modes is True:
            modes = self.modes
        elif modes is False:
            modes = None
        return ModeEmbedding(self.embedding, modes, masses=masses, mass_weight=mass_weight, dimensionless=dimensionless)

    def potential_expansion(self,
                            order=None,
                            *,
                            dimensionless=False,
                            embedding=None,
                            exclude_gradient=False,
                            **embedding_opts
                            ):
        if embedding is None:
            embedding = self.get_embedding(dimensionless=dimensionless, **embedding_opts)

        ders = self.potential.derivatives
        if exclude_gradient:
            ders = [0] + ders[1:]

        exp = ScalarExpansion(embedding, ders)
        if order is not None:
            exp = exp.get_terms(order)
        return exp
    def dipole_expansion(self,
                            order=None,
                            *,
                            expansion=None,
                            dimensionless=False,
                            embedding=None,
                            exclude_gradient=False,
                            **embedding_opts
                            ):
        if embedding is None:
            embedding = self.get_embedding(dimensionless=dimensionless, **embedding_opts)

        if expansion is None:
            expansion = self.dipole.derivatives

        exp = DipoleExpansion(embedding, expansion)
        if order is not None:
            exp = exp.get_terms(order, shared=1)
        return exp
    def get_potential_optmizing_transformation(self,
                                               order,
                                               *,
                                               dimensionless=False,
                                               embedding=None,
                                               exclude_gradient=True,
                                               **embedding_opts
                                               ):
        return nput.optimizing_transformation(
            self.potential_expansion(order,
                                     dimensionless=dimensionless,
                                     embedding=embedding,
                                     exclude_gradient=exclude_gradient,
                                     **embedding_opts
                                     ),
            order
        )

    def _get_ke_expansion(self, cls,
                          order=None,
                          *,
                          coords=None,
                          masses=None,
                          dimensionless=False,
                          embedding=None,
                          **embedding_opts
                          ):
        if embedding is None:
            embedding = self.get_embedding(masses=masses, dimensionless=dimensionless, **embedding_opts)

        exp = cls(embedding)
        if order is not None:
            exp = exp.get_terms(order, coords=coords)
        return exp
    def gmatrix_expansion(self,
                          order=None,
                          *,
                          coords=None,
                          dimensionless=False,
                          embedding=None,
                          **embedding_opts
                          ) -> "GMatrixExpansion|np.ndarray":
        return self._get_ke_expansion(GMatrixExpansion,
                                      coords=coords,
                                      order=order,
                                      dimensionless=dimensionless,
                                      embedding=embedding,
                                      **embedding_opts
                                      )
    def get_kinetic_optmizing_transformation(self,
                                             order,
                                             *,
                                             dimensionless=False,
                                             embedding=None,
                                             **embedding_opts
                                             ):
        g = self.gmatrix_expansion(dimensionless=dimensionless, embedding=embedding, **embedding_opts) #type:GMatrixExpansion
        G_r = g.get_terms(order)

        w = np.diag(G_r[0])
        # Q = [v]
        # R = [v.T]

        # Q0 = Q[0]
        # R0 = R[0]
        # G0 = G_r[0]
        Q = [np.eye(len(w))]
        R = [np.eye(len(w))]

        # w = w
        G_partial = None
        for o in range(1, order + 1):
            G_rem, G_partial = g.reexpress_G(G_r, Q, R,
                                             order=[o],
                                             # GR_expansion=G_partial,
                                             return_GR=True
                                             )
            # G_partial = G_partial[:-1]
            # w = w[np.newaxis]
            ggg = (
                          G_rem[-1] + G_rem[-1].transpose(1, 0, 2)
                   ) / 4
            # print(ggg - ggg.transpose(1, 0, 2))
            # raise ValueError("wtf")
            new_R = -ggg / (w[:, np.newaxis, np.newaxis] + w[np.newaxis, :, np.newaxis])
            # for i in range(o+1):
            #     new_R = np.tensordot(w, new_R)
            # print(w.shape)
            # print(new_R - np.transpose(new_R, (1, 0, 2)))
            # raise Exception(...)
            R = R + [new_R]

            print(new_R - new_R.transpose(1, 0, 2))

            Q = nput.inverse_transformation(R, [o], reverse_expansion=Q)
            print("="*50)
            G_rem2, G_partial = g.reexpress_G(G_r[:1], Q, R,
                                             order=[o],
                                             # GR_expansion=G_partial,
                                             return_GR=True
                                             )
            # print(G_rem[-1])
            print(G_rem2[-1] + G_rem2[-1].transpose(1, 0, 2))

            raise Exception("???")

        return Q, R

    def pseudopotential_expansion(self,
                                  order=None,
                                  *,
                                  dimensionless=False,
                                  embedding=None,
                                  **embedding_opts
                                  ):
        return self._get_ke_expansion(PseudopotentialExpansion,
                                      order=order,
                                      dimensionless=dimensionless,
                                      embedding=embedding,
                                      **embedding_opts
                                      )
    def coriolis_expansion(self,
                           order=None,
                           *,
                           dimensionless=False,
                           embedding=None,
                           **embedding_opts
                           ):
        return self._get_ke_expansion(CoriolisRotationExpansion,
                                      order=order,
                                      dimensionless=dimensionless,
                                      embedding=embedding,
                                      **embedding_opts
                                      )
    def watson_expansion(self,
                         order=None,
                         *,
                         dimensionless=False,
                         embedding=None,
                         **embedding_opts
                         ):
        return self._get_ke_expansion(ReciprocalInertiaExpansion,
                                      order=order,
                                      dimensionless=dimensionless,
                                      embedding=embedding,
                                      **embedding_opts
                                      )

class ScalarOperatorManager:
    def __init__(self, manager_or_derivs):
        self.manager = manager_or_derivs
        self._derivs = None
    @property
    def derivatives(self):
        if hasattr(self.manager, 'derivatives'):
            return self.manager.derivatives
        else:
            return self.manager

class PotentialManager(ScalarOperatorManager):
    def __init__(self, manager_or_derivs, modes:NormalModesManager):
        super().__init__(manager_or_derivs)
        modes = modes.modes
        if modes is not None:
            modes = modes.basis.to_new_modes().make_dimensionless()
        self.modes = modes

    @property
    def derivatives(self):
        if hasattr(self.manager, 'derivatives'):
            derivs = self.manager.derivatives
        else:
            derivs = self.manager
        return self.canonicalize_derivs(derivs)


class ScalarExpansion:
    """
    A helper class that can transform scalar derivatives from Cartesians (or Cartesian modes) to internals
    """
    def __init__(self,
                 embedding:ModeEmbedding,
                 derivs
                 ):
        self.embedding = embedding
        self.derivs = derivs

    def canonicalize_derivs(self, derivs, shared=0):
        if self.embedding.modes is None: return derivs
        # in_modes = False
        # for d in derivs:
        #     if not nput.is_zero(d):
        #         in_modes = d.shape[0] != d.shape[-1]
        #         if in_modes: break
        YX = self.embedding.mw_inverse()
        _ = []
        for d in derivs:
            if not nput.is_zero(d):
                for i,x in enumerate(d.shape[shared:]):
                    if x == YX.shape[-1]:
                        d = np.moveaxis(
                            np.tensordot(YX, d, axes=[-1, shared+i]),
                            0,
                            shared + i
                        )
            _.append(d)
        # if in_modes:
        #     QX = self.embedding.modes.coords_by_modes @ YX
        #     Qq = np.diag(1/np.sqrt(self.embedding.modes.freqs))
        #     _ = []
        #     for d in derivs:
        #         if not nput.is_zero(d):
        #             for j in range(d.ndim):
        #                 if d.shape[j] == QX.shape[1]:
        #                     d = np.tensordot(QX, d, axes=[1, j])
        #                 else:
        #                     d = np.tensordot(Qq, d, axes=[1, j])
        #                 d = np.moveaxis(d, 0, j)
        #             # # symmetrize d
        #             if symmetrization_mode is not None and symmetrization_mode != 'unhandled':
        #                 nput.symmetrize_array(d, symmetrization_mode=symmetrization_mode)
        #                 # n = 0
        #                 # t = 0
        #                 # for p in itertools.permutations(range(d.ndim-1)):
        #                 #     n += 1
        #                 #     t += d.transpose(p + (d.ndim-1,))
        #                 # d = t / n
        #         _.append(d)
        #     derivs = _
        # else:
        #     derivs = nput.tensor_reexpand([YX], derivs, len(derivs))
        return derivs

    @classmethod
    def check_mixed_expansion(self, derivs, shared=0):
        real_ders = [d for d in derivs if not nput.is_zero(d)]
        if any(d.shape[shared] != d.shape[-1] for d in real_ders):
            shp = real_ders[-1].shape
            mixed_axes = itut.counts(shp)[shp[-1]]
            return True, mixed_axes
        else:
            return False, None

    def get_scalar_transformation(self, order, transformation=None):
        base_tf = self.embedding.get_internals_by_cartesians(order)
        rev_tf = self.embedding.get_cartesians_by_internals(order)
        if transformation is not None:
            forward_derivs, reverse_derivs = MolecularHamiltonian.prep_transformation(transformation)
            base_tf = nput.tensor_reexpand(forward_derivs, base_tf)
            rev_tf = nput.tensor_reexpand(reverse_derivs, rev_tf)
        return base_tf, rev_tf

    def get_terms(self,
                  order=None,
                  *,
                  derivs=None,
                  transformation=None,
                  mixed_transformation=None,
                  mixed_derivative_handling_mode='high',
                  modes=None,
                  shared=0
                  ):
        if transformation is not None:
            forward_derivs, reverse_derivs = MolecularHamiltonian.prep_transformation(transformation)
            V = self.get_terms(order, derivs=derivs, shared=shared)
            return nput.tensor_reexpand(forward_derivs, V, order, axes=[-1, shared])

        if derivs is None:
            derivs = self.derivs
        if order is None:
            order = len(derivs)
        if len(derivs) < order:
            derivs = tuple(derivs) + (0,) * (order - len(derivs))

        derivs = self.canonicalize_derivs(derivs, shared=shared)
        is_mixed, num_base = self.check_mixed_expansion(derivs)

        zero_derivs = 0
        for d in derivs:
            if nput.is_numeric(d) and d == 0:
                d+=1
            else:
                break

        QX = self.embedding.get_cartesians_by_internals(order=order - zero_derivs)
        real_ders = [d for d in derivs if not nput.is_zero(d)]
        if self.embedding.modes is not None and (
                real_ders[0].shape[0] == self.embedding.modes.modes_by_coords.shape[1]
        ):
            XQ = self.embedding.modes.modes_by_coords
            QX = nput.tensor_reexpand(QX, [XQ], order - zero_derivs, axes=[-1, 0])
        else:
            YX = self.embedding.mw_inverse()
            QX = [np.tensordot(q, YX, axes=[-1, 0]) for q in QX]


        if is_mixed:
            canonical_derivs = nput.tensor_reexpand(
                QX[:num_base], derivs[:num_base], num_base,
                axes=[-1, -1]
            ) + [
                nput.tensor_reexpand(QX, [0] * (num_base-1) + [d], num_base, axes=[-1, -1])[-1]
                for d in derivs[num_base:]
            ]
            if mixed_transformation is None:
                if modes is None:
                    modes = self.embedding.modes
                if modes.is_cartesian:
                    if self.embedding.embedding.internals is not None:
                        mixed_transformation = nput.tensor_reexpand(
                            QX,
                            [modes.modes_by_coords]
                        )
            if mixed_transformation is None:
                return canonical_derivs
            else:
                tf_base = nput.tensor_reexpand(mixed_transformation, canonical_derivs, order, axes=[-1, -1])
                return [
                    nput.symmetrize_array(t, symmetrization_mode=mixed_derivative_handling_mode)
                    for t in tf_base
                ]
        else:
            # return TensorDerivativeConverter(QX, derivs).convert()
            return nput.tensor_reexpand(QX, derivs, order, axes=[-1, shared])

class DipoleExpansion(ScalarExpansion):
    def __init__(self,
                 embedding: ModeEmbedding,
                 derivs
                 ):
        ders = [
            np.moveaxis(d, -1, 0)
                if not nput.is_zero(d) and d.shape[-1] == 3 else
            d
            for d in derivs
        ]
        super().__init__(embedding, ders)
    def get_terms(self,
                  order=None,
                  *,
                  derivs=None,
                  transformation=None,
                  mixed_transformation=None,
                  mixed_derivative_handling_mode='high',
                  modes=None,
                  shared=0
                  ):
        if derivs is not None:
            derivs = [
                np.moveaxis(d, -1, 0)
                if not nput.is_zero(d) and d.shape[-1] == 3 else
                d
                for d in derivs
            ]
        else:
            derivs = self.derivs
        ref = derivs[0]
        derivs = derivs[1:]
        terms = super().get_terms(
            order=order,
            derivs=derivs,
            transformation=transformation,
            mixed_transformation=mixed_transformation,
            mixed_derivative_handling_mode=mixed_derivative_handling_mode,
            modes=modes,
            shared=shared
        )
        expansion = [ref] + terms
        new_exp = [
            np.moveaxis(d, 0, -1)
                if not nput.is_zero(d) and d.shape[0] == 3 else
            d
            for d in expansion
        ]
        return new_exp


class GMatrixExpansion:
    # I thought about having this be a generic "tensor product" expansion, but this is cleaner
    def __init__(self, embedding:ModeEmbedding):
        self.embedding = embedding

    def get_terms(self, order, coords=None, transformation=None):
        if transformation is not None:
            forward_derivs, reverse_derivs = MolecularHamiltonian.prep_transformation(transformation)
            return self.reexpress(order, forward_derivs, reverse_derivs)
        XQ = self.embedding.get_internals_by_cartesians(order=order+1, coords=coords)
        XG = nput.tensordot_deriv(XQ, XQ, order, axes=[-2, -2], shared=XQ[0].ndim-2)#, identical=False)

        if order > 0:
            QX = self.embedding.get_cartesians_by_internals(order=order+1, coords=coords)
            # XG = self.reexpress_G(XG, QX, XQ, order=order)
            XG = [XG[0]] + nput.tensor_reexpand(QX, XG[1:], order)
        return XG

    @classmethod
    def _dRGQ_partition_contrib(cls, partition, R, G):
        # print(partition)
        r1, r2, s = partition
        if s >= len(G): return 0
        if r1 >= len(R) or r2 >= len(R): return 0

        base_term = G[s]
        if isinstance(base_term, (int, float, np.integer, np.floating)) and base_term == 0:
            return 0

        r_perm_counter = 1
        r_perm_idx = []
        # g_perm_idx = []

        a = base_term.ndim - 2
        for r in [r1, r2]:
            d = R[r]
            if nput.is_zero(d): return 0

            if r == 0:
                base_term = np.moveaxis(base_term, a, -1)
            else:
                base_term = np.tensordot(base_term, d, axes=[a, 0])
                r_perm_idx.extend([r_perm_counter] * r)
                r_perm_counter -= 1
        if r2 > 0:
            # 1 1 -> (x, i, y, j) -> (-3 -> -2)
            base_term = np.moveaxis(base_term, -(r2 + 2), -2) # axis from the r1 contraction
        # if s > 0:
        g_perm_idx = ([1]*s) + ([0]*(r1+r2))

        if r1 > 0 or r2 > 0:
            # r1 and r2 need to be permuted to preserve symmetry, but all perms can overcount
            nterms = TensorDerivativeConverter.compute_partition_terms([p for p in partition[:2] if p > 0])
            r_perms, _ = UniquePermutations(r_perm_idx).permutations(return_indices=True)

            overcount = len(r_perms) / nterms
            base_term = base_term / overcount

            # if r1 != r2:
            base_term = base_term + np.moveaxis(base_term, -1, -2)


            # g indices need to be swapped into the r coords sometimes too
            g_perms, _ = UniquePermutations(g_perm_idx).permutations(return_indices=True)

            # take direct product of permutation indices
            perm_inds = []
            # print(overcount)
            padding = list(range(base_term.ndim - 2, base_term.ndim))
            for p in g_perms:
                for r in r_perms:
                    perm_inds.append(
                        list(p[:s]) + list(p[s:][r]) + padding
                    )

            base_term = sum(
                base_term.transpose(p)
                for p in perm_inds
            )

        return base_term

    @classmethod
    def _dRGQ_derivs(cls, R, G, o):
        # parts = IntegerPartitionPermutations(o, dim=3).get_partition_permutations(flatten=True)

        # print("="*50)
        total_cont = 0
        for g in range(o+1):
            rem = (o+1)-g
            # don't want to go smaller than half of rem, b.c. symmetry
            for r1 in range(rem//2, rem):
                r2 = o - (g + r1)
                # print("-"*10)
                # print(r1, r2, g)
                pc = cls._dRGQ_partition_contrib([r1, r2, g], R, G)
                # print(pc[0])
                total_cont += pc
        # print("_"*50)
        return total_cont

    @classmethod
    def reexpress_G(self,
                    G_expansion, forward_derivs, reverse_derivs,
                    order,
                    *,
                    GR_expansion=None,
                    return_GR=False
                    ):
        """
        Apply a coordinate transformation to the G-matrix

        :param forward_derivs:
        :param reverse_derivs:
        :param order:
        :return:
        """
        R = reverse_derivs
        Q = forward_derivs
        G = G_expansion

        if nput.is_numeric(order): order = list(range(order+1))
        m = max(order)

        G_R = GR_expansion
        if G_R is None:
            G_R = [self._dRGQ_derivs(R, G, o) for o in range(1, m+1)]

        if len(G_R) < m:
            G_R = G_R + [self._dRGQ_derivs(R, G, o) for o in range(len(G_R)+1, m+1)]

        rexpansion = nput.tensor_reexpand(Q,
                                          G_R,
                                          order=[o for o in order if o > 0],
                                          axes=[-1, 0]
                                          )
        G_Q = []
        n = 0
        G0 = None
        for o in order:
            if o == 0:
                if G0 is None:
                    G0 = reverse_derivs[0] @ G[0] @ reverse_derivs[0].T
                G_Q.append(G0)
            else:
                G_Q.append(rexpansion[n])
                n += 1

        if return_GR:
            return G_Q, G_R
        else:
            return G_Q

    def reexpress(self, order, forward_derivs, reverse_derivs):
        return self.reexpress_G(
            self.get_terms(order),
            forward_derivs,
            reverse_derivs,
            order=order
        )

class PseudopotentialExpansion:
    def __init__(self, g_expansion:"ModeEmbedding|GMatrixExpansion"):
        if not isinstance(g_expansion, GMatrixExpansion):
            g_expansion = GMatrixExpansion(g_expansion)
        self.g_expansion = g_expansion

    @classmethod
    def lambda_expansion(cls,
                        G_expansion, G_inv,
                        I0_expansion, I0_inv,
                        order):

        g_exp = [np.trace(t, axis1=-1, axis2=-2) for t in nput.tensordot_deriv(G_expansion[1:], G_inv, order)]
        i_exp = [np.trace(t, axis1=-1, axis2=-2) for t in nput.tensordot_deriv(I0_expansion[1:], I0_inv, order)]
        return [ti - tg for ti,tg in zip(i_exp, g_exp)]

    @classmethod
    def get_U(cls, G_expansion, I0_expansion):

        QG = G_expansion
        QI0 = I0_expansion
        n = len(QG) - 1
        order = n - 2

        QU = nput.matinv_deriv(QI0, n)
        QiG = nput.matinv_deriv(G_expansion, n)

        L = cls.lambda_expansion(
            QG, QiG,
            QI0, QU,
            order + 1
        )

        LxL = nput.tensorprod_deriv(L, L, order)

        inner_exp = [
            l1 + 1/4*l2
            for l1, l2 in zip(L[1:], LxL)
        ]
        G_contract = nput.tensordot_deriv(G_expansion, inner_exp, order, axes=[[0, 1], [0, 1]])
        QG_contract = [
            np.trace(g, axis1=-1, axis2=-2) for g in
            nput.tensordot_deriv(G_expansion[1:], L, order, axes=[2, 0])
        ]

        return [g1 + g2 for g1, g2 in zip(G_contract, QG_contract)]
        # return nput.scalarprod_deriv(Qig, t, order=order)

    def get_terms(self, order, transformation=None):
        if transformation is not None:
            forward_derivs, reverse_derivs = MolecularHamiltonian.prep_transformation(transformation)
            return self.reexpress(order, forward_derivs, reverse_derivs)

        # build setup terms, need to go 2 orders past the nominal order
        n = order + 2
        QG = self.g_expansion.get_terms(n)
        QI0 = self.g_expansion.embedding.get_inertia_tensor_expansion(order=n)

        return self.get_U(QG, QI0)

    def reexpress(self, order, forward_derivs, reverse_derivs):
        """
        Apply a coordinate transformation to the G-matrix

        :param forward_derivs:
        :param reverse_derivs:
        :param order:
        :return:
        """

        # build setup terms, need to go 2 orders past the nominal order
        n = order + 2
        QG = self.g_expansion.reexpress(n, forward_derivs, reverse_derivs)
        QI0 = nput.tensor_reexpand(
            forward_derivs,
            self.g_expansion.embedding.get_inertia_tensor_expansion(order=n),
            n
        )

        return self.get_U(QG, QI0)

class ZetaExpansion:
    def __init__(self, embedding:ModeEmbedding):
        self.embedding = embedding
    def get_terms(self, order, transformation=None):

        QY = self.embedding.get_internals_by_cartesians(order=order)
        if transformation is not None:
            forward_derivs, reverse_derivs = MolecularHamiltonian.prep_transformation(transformation)
            QY = nput.tensor_reexpand(forward_derivs, QY, order)
        QY = [
            np.moveaxis(yq, -1, 0) if not nput.is_zero(yq) else 0
            for yq in QY
        ]

        B_e, eigs = self.embedding.get_inertial_frame()

        _ = []
        for J in QY:
            J = J.reshape(J.shape[:-2] + (-1, 3))
            J = np.tensordot(eigs, J, axes=[1, 2])
            _.append(np.moveaxis(J, -2, -1))
        QY = _

        QYQY = [
            np.sum(qyqy, axis=-1)
            for qyqy in nput.tensorprod_deriv(QY, QY, axes=[[0, 1], [0, 1]], order=order)
        ]

        zeta_expansion = []
        for qyqy in QYQY:
            for a in range(3):
                b = (a + 1) % 3
                c = (a + 2) % 3
                if b % 2 == 1:
                    c, b = b, c
                zeta_expansion.append(qyqy[..., b, :, c] - qyqy[..., c, :, b])

        return zeta_expansion

class ReciprocalInertiaExpansion:
    """
    Pulled from Watson
    """
    def __init__(self, embedding:ModeEmbedding):
        self.embedding = embedding

    def get_terms(self, order, transformation=None):
        I0, a = self.embedding.get_inertia_tensor_expansion(order=1)
        if transformation is not None:
            forward_derivs, reverse_derivs = MolecularHamiltonian.prep_transformation(transformation)
            a = nput.tensor_reexpand(forward_derivs, [a], 1)[0]
        u = np.linalg.inv(I0)

        expansion = [u]
        for _ in range(order):
            prev = expansion[-1]
            new = np.tensordot(
                np.tensordot(a, prev, axes=[-1, -1]),
                u,
                axes=[-1, 0]
            )
            expansion.append(new)

        return expansion

class CoriolisRotationExpansion:
    """
    Provides an expansion of the Coriolis rotation operator
    """
    def __init__(self, embedding:ModeEmbedding):
        self.embedding = embedding
        self.zetas = ZetaExpansion(self.embedding)
        self.us = ReciprocalInertiaExpansion(self.embedding)

    def get_terms(self, order, transformation=None):
        Z = self.zetas.get_terms(order, transformation=transformation)
        ZZ = nput.tensorprod_deriv(Z, Z, order, axes=[1, 1]) # x/y/z shared
        u = self.us.get_terms(order, transformation=transformation)

        return nput.tensordot_deriv(u, ZZ, order, axes=[1, 0])




import numpy as np, scipy.linalg as slag, collections

from McUtils.Data import AtomData, UnitsData
import McUtils.Numputils as nput

from .MixtureModes import MixtureModes

__all__ = [
    "NormalModes",
    "ReactionPathModes"
]

__reload_hook__ = [".MixtureModes"]

class NormalModes(MixtureModes):
    name="NormalModes"

    def __init__(self,
                 basis,
                 coeffs,
                 freqs=None,
                 origin=None,
                 masses=None,
                 inverse=None,
                 name=None,
                 mass_weighted=False,
                 frequency_scaled=False,
                 g_matrix=None
                 ):
        super().__init__(
            basis,
            coeffs,
            freqs=freqs,
            origin=origin,
            masses=masses,
            inverse=inverse,
            mass_weighted=mass_weighted,
            frequency_scaled=frequency_scaled,
            g_matrix=g_matrix,
            name=name
        )

    ModeData = collections.namedtuple("ModeData", ['freqs', 'modes', 'inverse'])
    default_zero_freq_cutoff = 1.0e-4 # 20 wavenumbers...
    @classmethod
    def get_normal_modes(cls,
                         f_matrix,
                         mass_spec,
                         # mass_units="AtomicMassUnits",
                         remove_transrot=True,
                         dimensionless=False,
                         mass_weighted=None,
                         zero_freq_cutoff=None,
                         return_gmatrix=False,
                         projector=None
                         ):

        f_matrix = np.asanyarray(f_matrix)
        base_shape = f_matrix.shape[:-2]
        f_matrix = f_matrix.reshape((-1,) + f_matrix.shape[-2:])
        if isinstance(mass_spec[0], str):  # atoms were supplied
            mass_spec = np.array([AtomData[a, "Mass"] for a in mass_spec]) * UnitsData.convert(
                "AtomicMassUnits",
                "AtomicUnitOfMass"
            )
        mass_spec = np.asanyarray(mass_spec)

        if mass_spec.ndim == 1:
            if mass_weighted is None: mass_weighted = dimensionless # generally do the right thing for Cartesians
            mass_spec = np.broadcast_to(mass_spec[:, np.newaxis], (len(mass_spec), 3)).flatten()
            mass_spec = np.diag(1 / mass_spec)

        mass_spec = np.reshape(mass_spec, (-1,) + mass_spec.shape[-2:])
        if mass_spec.shape[0] < f_matrix.shape[0]:
            mass_spec = np.broadcast_to(mass_spec, f_matrix.shape)

        gi12 = nput.fractional_power(mass_spec, -1 / 2)
        g12 = nput.fractional_power(mass_spec, 1 / 2)
        if projector is None:
            if f_matrix.shape[0] == 1:
                freq2, modes = slag.eigh(f_matrix[0], mass_spec[0], type=3)
                V = gi12[0] @ modes
                V = V[np.newaxis]
                freq2 = freq2[np.newaxis]
                modes = modes[np.newaxis]
            else:
                mw_F = g12 @ f_matrix @ np.moveaxis(g12, -1, -2)
                freq2, V = np.linalg.eigh(mw_F)
                modes = g12 @ V
        else:
            pG12 = projector @ g12
            mw_F = pG12 @ f_matrix @ np.moveaxis(pG12, -1, -2)
            freq2, V = np.linalg.eigh(mw_F)
            modes = g12 @ V

        # modes = g12 @ V
        # therefore, inv = V.T @ gi12
        # and modes[:, nz] = g12 @ V[:, nz]; inv[nz, :] = (V.T)[nz, :] @ gi12 (this is only the _left_ inverse)

        blocks = False
        if remove_transrot:
            if zero_freq_cutoff is None:
                zero_freq_cutoff = cls.default_zero_freq_cutoff
            if zero_freq_cutoff > 0:
                nonzero = np.abs(freq2) >= zero_freq_cutoff**2
            else:
                nonzero = np.full(freq2.shape, True)

            if gi12.shape[0] != modes.shape[0]:
                gi12 = np.broadcast_to(gi12, f_matrix.shape)

            # need to check if shape is good
            nz_counts = np.sum(nonzero, axis=1)
            blocks = len(freq2) == 1 or np.sum(np.abs(np.diff(nz_counts))) != 0
            freq2_ = []
            modes_ = []
            inv_ = []
            for nz2, f2, m2, v2, gi in zip(nonzero, freq2, modes, V, gi12):
                freq2_.append(f2[nz2])
                modes_.append(m2[:, nz2])
                inv_.append(v2.T[nz2, :] @ gi)
            freq2 = freq2_
            modes = modes_
            inv = inv_
            del freq2_
            del modes_
            del inv_
            if not blocks:
                freq2 = np.array(freq2)
                modes = np.array(modes)
                inv = np.array(inv)

            # else:
            #     if nz_counts[0] == 0:
            #         inv = np.moveaxis(V, -1, -2) @ gi12
            #     else:
            #         nz_spec = [np.where(nz)[0] for nz in nonzero]
            #         freq2 = freq2[nonzero].reshape((modes.shape[0], nz_counts[0]))
            #         print(nz_spec, modes.shape)
            #         modes = np.moveaxis(
            #             nput.vector_take(
            #                 np.moveaxis(modes, -1, -2),
            #                 nz_spec,
            #                 shared=1
            #             ),
            #             -1, -2
            #         )
            #         inv = np.moveaxis(
            #                 nput.vector_take(
            #                     np.moveaxis(V, 1, 0),
            #                     nz_spec,
            #                     shared=1
            #                 )
            #         print(modes.shape, inv.shape)
        else:
            inv = np.moveaxis(V, -1, -2) @ gi12
        # else:
        #     if np.linalg.det(modes) == 1 and np.allclose(modes.T @ modes, np.eyelen(modes)):
        #         inv = modes.T
        #     else:
        #         inv = np.linalg.inv(modes)

        if blocks:
            freqs = [np.sign(f2) * np.sqrt(np.abs(f2)) for f2 in freq2]
        else:
            freqs = np.sign(freq2) * np.sqrt(np.abs(freq2))

        if mass_weighted:
            if blocks:
                modes_ = []
                inv_ = []
                for m,i,g,gi in zip(modes, inv, g12, gi12):
                    modes_.append(gi @ m)
                    inv_.append(i @ g)
                modes = modes_
                inv = inv_
            else:
                # gi12 = slag.fractional_matrix_power(mass_spec, -1 / 2)
                # g12 = slag.fractional_matrix_power(mass_spec, 1 / 2)
                inv = inv @ g12
                modes = gi12 @ modes
        if dimensionless:
            if blocks:
                modes_ = []
                inv_ = []
                for f,m,i in zip(freqs, modes, inv):
                    weighting = np.sqrt(np.abs(f))
                    modes_.append(m / weighting[np.newaxis, :])
                    inv_.append(i * weighting[:, np.newaxis])
                modes = modes_
                inv = inv_
            else:
                weighting = np.sqrt(np.abs(freqs))
                modes = modes / weighting[:, np.newaxis, :]
                inv = inv * weighting[:, :, np.newaxis]
        # if mode == 'reasonable':
        # modes, inv = inv.T, modes.T

        if blocks:
            modes_ = []
            inv_ = []
            for f, m, i in zip(freqs, modes, inv):
                sorting = np.argsort(f)
                modes_.append(m[:, sorting])
                inv_.append(i[sorting, :])
            modes = modes_
            inv = inv_
            modes, inv = [np.moveaxis(i, -1, -2) for i in inv], [np.moveaxis(m, -1, -2) for m in modes]
            # TODO: iterative reshaping
            if len(base_shape) == 0:
                freqs = freqs[0]
                modes = modes[0]
                inv = inv[0]
        else:
            modes, inv = np.moveaxis(inv, -1, -2), np.moveaxis(modes, -1, -2)
            freqs = np.reshape(freqs, base_shape + freqs.shape[1:])
            modes = np.reshape(modes, base_shape + modes.shape[1:])
            inv = np.reshape(inv, base_shape + inv.shape[1:])

        mode_data = cls.ModeData(freqs, modes, inv)
        if return_gmatrix:
            mass_spec = mass_spec.reshape(base_shape + mass_spec.shape[-2:])
            return mode_data, mass_spec
        else:
            return mode_data

    @classmethod
    def from_fg(cls,
                basis,
                f_matrix,
                mass_spec,
                remove_transrot=True,
                dimensionless=False,
                zero_freq_cutoff=None,
                mass_weighted=None,
                origin=None,
                projector=None,
                **opts
                ):
        """
        Generates normal modes from the specified F and G matrices

        :param basis:
        :param f_matrix: second derivatives of the potential
        :param mass_spec:
        :param mass_units:
        :param remove_transrot:
        :param opts:
        :return:
        """

        (freqs, modes, inv), g_matrix = cls.get_normal_modes(f_matrix, mass_spec,
                                                             remove_transrot=remove_transrot,
                                                             dimensionless=dimensionless,
                                                             zero_freq_cutoff=zero_freq_cutoff,
                                                             mass_weighted=mass_weighted,
                                                             return_gmatrix=True,
                                                             projector=projector
                                                             )

        mass_weighted = mass_weighted or dimensionless

        if mass_weighted and origin is not None:
            origin = np.asanyarray(origin)
            origin = (origin.flatten() @ nput.fractional_power(g_matrix, -1/2)).reshape(origin.shape)

        return cls(basis, modes, inverse=inv, freqs=freqs,
                   mass_weighted=mass_weighted,
                   frequency_scaled=dimensionless,
                   g_matrix=g_matrix,
                   origin=origin,
                   **opts
                   )

    default_projected_zero_freq_cutoff = None
    @classmethod
    def from_molecule(cls, mol,
                      dimensionless=False,
                      use_internals=None,
                      potential_derivatives=None,
                      project_transrot=True,
                      zero_freq_cutoff=None,
                      masses=None,
                      energy_evaluator=None,
                      **opts
                      ):
        from ..Molecools import Molecule

        if (
                not project_transrot
                and (
                    not use_internals
                    or (use_internals is None and mol.internals is None)
                )
                and mol.normal_modes.get_modes(quiet=True) is not None
        ):
            return mol.normal_modes.modes.basis.to_new_modes()

        mol: Molecule
        if use_internals is None:
            use_internals = mol.internal_coordinates is not None
        og_potential_derivatives = potential_derivatives
        if potential_derivatives is None:
            potential_derivatives = mol.potential_derivatives
            if potential_derivatives is None:
                if (
                        energy_evaluator is not None or
                        mol.energy_evaluator is not None
                ):
                    potential_derivatives = mol.calculate_energy(evaluator=energy_evaluator, order=2)[1:]
                else:
                    modes = mol.normal_modes.get_modes(quiet=True)
                    if modes is not None:
                        new: NormalModes = modes.basis.to_new_modes().remove_mass_weighting()
                        f = new.modes_by_coords @ np.diag(new.freqs**2) @ new.modes_by_coords.T
                        potential_derivatives = [0, f]
        if potential_derivatives is None:
            raise ValueError(f"{mol} has no potential derivatives or existing normal modes")

        if use_internals:
            hess = nput.tensor_reexpand(
                mol.get_cartesians_by_internals(1, strip_embedding=True, reembed=True),
                [0, potential_derivatives[1]],
                order=2
            )[1]
            if zero_freq_cutoff is None:
                zero_freq_cutoff = cls.default_projected_zero_freq_cutoff
                if zero_freq_cutoff is None:
                    zero_freq_cutoff = 1 * UnitsData.convert("Wavenumbers", "Hartrees")
            return cls.from_fg(
                mol.internal_coordinates.system,
                hess,
                mol.get_gmatrix(masses=masses),
                dimensionless=dimensionless,
                origin=mol.get_internals(strip_embedding=True),
                zero_freq_cutoff=zero_freq_cutoff,
                **opts
            )
        else:
            if (
                    not project_transrot
                    and mol.normal_modes.modes is not None
                    and og_potential_derivatives is None
            ):
                return mol.normal_modes.modes.basis.to_new_modes()

            if masses is None:
                masses = mol.atomic_masses
            hess = potential_derivatives[1]
            if project_transrot:
                proj = mol.get_translation_rotation_projector(mass_weighted=True)
                mw = np.diag(np.repeat(1 / np.sqrt(masses), 3))
                mw_inv = np.diag(np.repeat(np.sqrt(masses), 3))
                proj = mw_inv @ proj @ mw
                hess = proj @ hess @ proj.T
                if zero_freq_cutoff is None:
                    zero_freq_cutoff = cls.default_projected_zero_freq_cutoff
                    if zero_freq_cutoff is None:
                        zero_freq_cutoff = 1 * UnitsData.convert("Wavenumbers", "Hartrees")

            return cls.from_fg(
                mol.coords.system,
                hess,
                masses,
                dimensionless=dimensionless,
                masses=masses,
                origin=mol.coords,
                zero_freq_cutoff=zero_freq_cutoff,
                **opts
            )

    def get_reaction_path_modes(self,
                                grad: np.ndarray,
                                zero_gradient_cutoff=None,
                                return_status=False,
                                **kwargs
                                ):
        return ReactionPathModes.from_modes_and_grad(self, grad,
                                                     zero_gradient_cutoff=zero_gradient_cutoff,
                                                     return_status=return_status,
                                                     **kwargs
                                                     )

class ReactionPathModes(NormalModes):

    zero_gradient_cutoff = 2.5e-5#5 cm-1
    @classmethod
    def get_rp_modes(cls,
                     gradient,
                     f_matrix,
                     mass_spec,
                     # mass_units="AtomicMassUnits",
                     remove_transrot=True,
                     dimensionless=False,
                     mass_weighted=None,
                     zero_freq_cutoff=None,
                     return_gmatrix=False,
                     projector=None,
                     zero_gradient_cutoff=None,
                     use_max_gradient_cutoff=True,
                     return_indices=False
                     ):
        if zero_gradient_cutoff is None:
            zero_gradient_cutoff = cls.zero_gradient_cutoff

        gradient = np.asanyarray(gradient)
        f_matrix = np.asanyarray(f_matrix)
        base_shape = f_matrix.shape[:-2]
        if f_matrix.ndim > 2:
            if gradient.ndim == 1:
                gradient = gradient[np.newaxis, :, np.newaxis]
            elif gradient.shape[:-2] == base_shape:
                gradient = gradient.reshape((-1,) + gradient.shape[-2:])
            elif gradient.shape[:-1] == base_shape:
                gradient = gradient.reshape((-1, gradient.shape[-1], 1))
            else:
                raise ValueError("can't make gradient of shape {} work with F-matrix of shape {}".format(
                    gradient.shape,
                    f_matrix.shape
                ))
        else:
            if gradient.ndim == 1:
                gradient = gradient[np.newaxis, :, np.newaxis]
            else:
                gradient = gradient[np.newaxis]

        f_matrix = f_matrix.reshape((-1,) + f_matrix.shape[-2:])
        if projector is not None:
            projector = np.asanyarray(projector)
            projector = projector.reshape((-1,) + projector.shape[-2:])

        if isinstance(mass_spec[0], str):  # atoms were supplied
            mass_spec = np.array([AtomData[a, "Mass"] for a in mass_spec]) * UnitsData.convert(
                "AtomicMassUnits",
                "AtomicUnitOfMass"
            )
        mass_spec = np.asanyarray(mass_spec)

        og_mass = mass_spec

        if mass_spec.ndim == 1:
            if mass_weighted is None: mass_weighted = dimensionless # generally do the right thing for Cartesians
            mass_spec = np.broadcast_to(mass_spec[:, np.newaxis], (len(mass_spec), 3)).flatten()
            mass_spec = np.diag(1 / mass_spec)

        mass_spec = np.reshape(mass_spec, (-1,) + mass_spec.shape[-2:])
        if mass_spec.shape[0] < f_matrix.shape[0]:
            mass_spec = np.broadcast_to(mass_spec, f_matrix.shape)

        gi12 = nput.fractional_power(mass_spec, -1 / 2)
        g12 = nput.fractional_power(mass_spec, 1 / 2)
        raw_grad = gradient
        gradient = g12 @ gradient

        grad_norm = nput.vec_norms(gradient, axis=1)
        if zero_gradient_cutoff is not None:
            if use_max_gradient_cutoff:
                regular_mode_pos = np.where(np.max(np.abs(raw_grad), axis=1) < zero_gradient_cutoff)
            else:
                regular_mode_pos = np.where(grad_norm < zero_gradient_cutoff)
            if len(regular_mode_pos) > 0:
                regular_mode_pos = regular_mode_pos[0]
            rem_pos = np.delete(np.arange(len(grad_norm)), regular_mode_pos)
        else:
            regular_mode_pos = []
            rem_pos = np.arange(len(f_matrix))

        if len(regular_mode_pos) > 0:

            reg_modes = cls.get_normal_modes(
                f_matrix[regular_mode_pos,],
                og_mass,
                # mass_units="AtomicMassUnits",
                remove_transrot=remove_transrot,
                dimensionless=dimensionless,
                mass_weighted=mass_weighted,
                zero_freq_cutoff=zero_freq_cutoff,
                # return_gmatrix=True,
                projector=projector[regular_mode_pos,] if projector is not None else projector
            )
        else:
            reg_modes = None

        # if zero_gradient_cutoff is not None and grad_norm < zero_gradient_cutoff:
        #     return None


        gradient = nput.vec_normalize(gradient[rem_pos,], norms=grad_norm[rem_pos,], axis=1)
        grad_projector = nput.orthogonal_projection_matrix(gradient, orthonormal=True)
        if projector is not None:
            projector = grad_projector @ projector[rem_pos,] @ grad_projector
        else:
            projector = grad_projector

        if len(rem_pos) > 0:
            f_matrix = f_matrix[rem_pos,]
            (freqs, modes, inv), g_matrix = cls.get_normal_modes(
                f_matrix,
                og_mass,
                # mass_units="AtomicMassUnits",
                remove_transrot=remove_transrot,
                dimensionless=dimensionless,
                mass_weighted=mass_weighted,
                zero_freq_cutoff=zero_freq_cutoff,
                return_gmatrix=True,
                projector=projector
            )


            gi12 = gi12[rem_pos,]
            g12 = g12[rem_pos,]
            gradient_inv = np.moveaxis(gradient, -1, -2)
            f_grad = gradient_inv @ f_matrix @ np.moveaxis(gradient_inv, -1, -2)
            grad_g = np.moveaxis(gradient, -1, -2) @ g_matrix @ gradient
            if not mass_weighted:
                gradient = gi12 @ gradient
                gradient_inv = gradient_inv @ g12

            if not remove_transrot:
                # if isinstance(freqs, list):
                    freqs_ = []
                    modes_ = []
                    inv_ = []
                    for g,f,m,i in zip(gradient, freqs, modes, inv):
                        zero_pos = np.where(np.abs(f) < 1e-6)[0] # projected out
                        overlaps = (i[zero_pos, :] @ g)**2
                        max_sim_pos = np.argmax(overlaps)
                        # print(max_sim_pos)
                        del_idx = zero_pos[max_sim_pos]
                        freqs_.append(np.delete(f, del_idx))
                        modes_.append(np.delete(m, del_idx, axis=-1))
                        inv_.append(np.delete(i, del_idx, axis=-2))
                    if isinstance(freqs, list):
                        freqs, modes, inv = freqs_, modes_, inv_
                    else:
                        freqs, modes, inv = np.array(freqs_), np.array(modes_), np.array(inv_)

            rows, cols = np.diag_indices(grad_g.shape[-1])
            loc_2 = f_grad[..., rows, cols] * grad_g[..., rows, cols]
            freq = np.sign(loc_2) * np.sqrt(np.abs(loc_2))

            if isinstance(freqs, list):
                freqs = [
                    np.concatenate([f1, f2], axis=0)
                    for f1, f2 in zip(freq, freqs)
                ]

                modes = [
                    np.concatenate([g, m], axis=1)
                    for g,m in zip(gradient, modes)
                ]

                inv = [
                    np.concatenate([gi, i], axis=0)
                    for gi,i in zip(gradient_inv, inv)
                ]

                if reg_modes is not None:
                    reg_freqs, reg_modes, reg_inv = reg_modes
                    full_freqs:list[np.ndarray] = [None] * (len(reg_freqs) + len(freqs))
                    full_modes:list[np.ndarray] = [None] * (len(reg_freqs) + len(freqs))
                    full_inv:list[np.ndarray] = [None] * (len(reg_freqs) + len(freqs))
                    for r,f,m,i in zip(regular_mode_pos, reg_freqs, reg_modes, reg_inv):
                        full_freqs[r] = f
                        full_modes[r] = m
                        full_inv[r] = i
                    for r,f,m,i in zip(rem_pos, freqs, modes, inv):
                        full_freqs[r] = f
                        full_modes[r] = m
                        full_inv[r] = i

                    freqs, modes, inv = full_freqs, full_modes, full_inv

                if len(base_shape) == 0:
                    freqs = freqs[0]
                    modes = modes[0]
                    inv = inv[0]
                    mass_spec = mass_spec[0]
                # elif len(base_shape) > 1:
                #     ...
            else:
                freqs = np.concatenate([freq, freqs], axis=1)

                modes = np.concatenate(
                    [
                        gradient,
                        modes
                    ],
                    axis=2
                )

                inv = np.concatenate(
                    [
                        gradient_inv,
                        inv
                    ],
                    axis=1
                )

                if reg_modes is not None:
                    reg_freqs, reg_modes, reg_inv = reg_modes
                    full_freqs:np.ndarray = np.empty(
                        (len(reg_freqs) + len(freqs),) + freqs.shape[1:],
                        dtype=freqs.dtype
                    )
                    full_freqs[regular_mode_pos,] = reg_freqs
                    full_freqs[rem_pos,] = freqs

                    full_modes:np.ndarray = np.empty(
                        (len(reg_modes) + len(modes),) + modes.shape[1:],
                        dtype=modes.dtype
                    )
                    full_modes[regular_mode_pos,] = reg_modes
                    full_modes[rem_pos,] = modes


                    full_inv:np.ndarray = np.empty(
                        (len(reg_inv) + len(inv),) + inv.shape[1:],
                        dtype=inv.dtype
                    )
                    full_inv[regular_mode_pos,] = reg_inv
                    full_inv[rem_pos,] = inv

                    freqs, modes, inv = full_freqs, full_modes, full_inv

                freqs = freqs.reshape(base_shape + freqs.shape[1:])
                modes = modes.reshape(base_shape + modes.shape[1:])
                inv = inv.reshape(base_shape + inv.shape[1:])
                mass_spec = mass_spec.reshape(base_shape + mass_spec.shape[1:])

        else:
            freqs, modes, inv = reg_modes

            if isinstance(freqs, list):
                if len(base_shape) == 0:
                    freqs = freqs[0]
                    modes = modes[0]
                    inv = inv[0]
                    mass_spec = mass_spec[0]
            else:
                freqs = freqs.reshape(base_shape + freqs.shape[1:])
                modes = modes.reshape(base_shape + modes.shape[1:])
                inv = inv.reshape(base_shape + inv.shape[1:])
                mass_spec = mass_spec.reshape(base_shape + mass_spec.shape[1:])

        mode_data = cls.ModeData(freqs, modes, inv)
        if return_gmatrix or return_indices:
            res = (mode_data,)
            if return_gmatrix:
                res = res + (mass_spec,)
            if return_indices:
                res = res + ((regular_mode_pos, rem_pos),)
            return res
        else:
            return mode_data

    @classmethod
    def from_grad_fg(
            cls,
            basis,
            gradient,
            f_matrix,
            mass_spec,
            remove_transrot=True,
            dimensionless=False,
            zero_freq_cutoff=None,
            mass_weighted=None,
            origin=None,
            projector=None,
            zero_gradient_cutoff=None,
            return_status=False,
            **opts
    ):
        """
        Generates normal modes from the specified F and G matrices

        :param basis:
        :param f_matrix: second derivatives of the potential
        :param mass_spec:
        :param mass_units:
        :param remove_transrot:
        :param opts:
        :return:
        """

        mode_data = cls.get_rp_modes(
            gradient, f_matrix, mass_spec,
            remove_transrot=remove_transrot,
            dimensionless=dimensionless,
            zero_freq_cutoff=zero_freq_cutoff,
            mass_weighted=mass_weighted,
            return_gmatrix=True,
            projector=projector,
            zero_gradient_cutoff=zero_gradient_cutoff,
            return_indices=True
        )
        if mode_data is None:
            return None

        (freqs, modes, inv), g_matrix, inds = mode_data

        mass_weighted = mass_weighted or dimensionless

        if mass_weighted and origin is not None:
            origin = np.asanyarray(origin)
            origin = (origin.flatten() @ nput.fractional_power(g_matrix, -1 / 2)).reshape(origin.shape)

        new = cls(basis, modes, inverse=inv, freqs=freqs,
                   mass_weighted=mass_weighted,
                   frequency_scaled=dimensionless,
                   g_matrix=g_matrix,
                   origin=origin,
                   **opts
                   )
        if return_status:
            reg, rp = inds
            return new, len(rp) > 0
        else:
            return new

    @classmethod
    def from_molecule(cls, mol,
                      dimensionless=False,
                      use_internals=None,
                      potential_derivatives=None,
                      project_transrot=True,
                      zero_freq_cutoff=None,
                      masses=None,
                      zero_gradient_cutoff=None,
                      return_status=False,
                      **opts
                      ):
        from ..Molecools import Molecule
        #TODO: cut down on duplication

        mol = mol  # type:Molecule
        if use_internals is None:
            use_internals = mol.internal_coordinates is not None
        if potential_derivatives is None:
            potential_derivatives = mol.potential_derivatives
            if potential_derivatives is None and mol.energy_evaluator is not None:
                potential_derivatives = mol.calculate_energy(order=2)[1:]
        if use_internals:
            hess = nput.tensor_reexpand(
                mol.get_cartesians_by_internals(1, strip_embedding=True, reembed=True),
                [0, potential_derivatives[1]],
                order=2
            )[1]
            if zero_freq_cutoff is None:
                zero_freq_cutoff = cls.default_projected_zero_freq_cutoff
                if zero_freq_cutoff is None:
                    zero_freq_cutoff = 1 * UnitsData.convert("Wavenumbers", "Hartrees")
            return cls.from_grad_fg(
                mol.internal_coordinates.system,
                potential_derivatives[0],
                hess,
                mol.get_gmatrix(masses=masses),
                dimensionless=dimensionless,
                origin=mol.get_internals(strip_embedding=True),
                zero_freq_cutoff=zero_freq_cutoff,
                zero_gradient_cutoff=zero_gradient_cutoff,
                return_status=return_status,
                **opts
            )
        else:
            # if not project_transrot and mol.normal_modes.modes is not None:
            #     return mol.normal_modes.modes.basis.to_new_modes()

            if masses is None:
                masses = mol.atomic_masses
            hess = potential_derivatives[1]
            if project_transrot:
                proj = mol.get_translation_rotation_projector(mass_weighted=True)
                mw = np.diag(np.repeat(1 / np.sqrt(masses), 3))
                mw_inv = np.diag(np.repeat(np.sqrt(masses), 3))
                proj = mw_inv @ proj @ mw
                hess = proj @ hess @ proj.T
                if zero_freq_cutoff is None:
                    zero_freq_cutoff = cls.default_projected_zero_freq_cutoff
                    if zero_freq_cutoff is None:
                        zero_freq_cutoff = 1 * UnitsData.convert("Wavenumbers", "Hartrees")

            return cls.from_grad_fg(
                mol.coords.system,
                potential_derivatives[0],
                hess,
                masses,
                dimensionless=dimensionless,
                masses=masses,
                origin=mol.coords,
                zero_freq_cutoff=zero_freq_cutoff,
                zero_gradient_cutoff=zero_gradient_cutoff,
                return_status=return_status,
                **opts
            )

    @classmethod
    def from_modes_and_grad(cls,
                            modes:MixtureModes,
                            grad:np.ndarray,
                            zero_gradient_cutoff=None,
                            use_max_gradient_cutoff=True,
                            return_status=False,
                            mass_weighted=None,
                            **projection_opts
                            ):
        g, g12, gi12 = modes.compute_gmatrix('coords', return_fractional=True)
        new_grad = g12 @ grad

        grad_norm = nput.vec_norms(new_grad)
        if zero_gradient_cutoff is not None:
            if use_max_gradient_cutoff:
                stat = np.max(np.abs(grad)) > zero_gradient_cutoff
            else:
                stat = grad_norm > zero_gradient_cutoff
        else:
            stat = True
        if stat:

            gradient_mw = new_grad[:, np.newaxis] / grad_norm
            gradient_inv_mw = np.moveaxis(gradient_mw, -2, -1)

            gradient = gi12 @ gradient_mw
            gradient_inv = gradient_inv_mw @ g12
            old_modes = modes.remove_mass_weighting()

            # print(1 / grad_norm, grad_norm)
            new_modes = old_modes.localize(
                projections=[gradient],
                orthogonal_projection=True,
                allow_mode_mixing=True,
                **projection_opts
            )

            if len(new_modes.freqs) == len(old_modes.freqs):
                print(
                    new_modes.local_freqs * UnitsData.hartrees_to_wavenumbers,
                    old_modes.local_freqs * UnitsData.hartrees_to_wavenumbers
                )
                raise ValueError("projection didn't change subspace size")

            # import McUtils.Formatters as mfmt
            #
            # print(mfmt.TableFormatter("{:.0f}").format(
            #     old_modes.local_hessian * UnitsData.hartrees_to_wavenumbers
            # ))
            #
            # print("-"*20)
            #
            # print(mfmt.TableFormatter("{:.0f}").format(
            #     new_modes.local_hessian * UnitsData.hartrees_to_wavenumbers
            # ))
            #
            # raise Exception(...)

            if mass_weighted is None:
                mass_weighted = modes.mass_weighted

            f_grad = gradient_inv @ old_modes.compute_hessian('coords') @ np.moveaxis(gradient_inv, -1, -2)
            if mass_weighted:
                new_modes = new_modes.make_mass_weighted()
                gradient = gradient_mw
                gradient_inv = gradient_inv_mw

            modes_by_coords = np.concatenate(
                [
                    gradient,
                    new_modes.modes_by_coords
                ],
                axis=1
            )
            coords_by_modes = np.concatenate(
                [
                    gradient_inv,
                    new_modes.coords_by_modes
                ],
                axis=0
            )

            modes = cls(
                old_modes.basis,
                modes_by_coords,
                inverse=coords_by_modes,
                origin=new_modes.origin,
                masses=new_modes.masses,
                freqs=np.concatenate([f_grad.flatten(), new_modes.freqs]),
                mass_weighted=mass_weighted
            )

        if return_status:
            return modes, stat
        else:
            return modes


import numpy as np
import scipy.linalg
from McUtils.Data import UnitsData
import McUtils.Devutils as dev
import McUtils.Numputils as nput
import McUtils.Coordinerds as coordops
from .StateSpaces import AbstractStateSpace, BasisStateSpace, BraKetSpace
from .Representations import ExpansionRepresentation
from .HarmonicOscillator import HarmonicOscillatorProductBasis
from ..Wavefun import MatrixWavefunctions
from . import Util as util

__all__ = [
    "LocalHarmonicModel",
    "CustomLocalHarmonicModel",
    "TaborCHModel"
]

class LocalHarmonicModel:
    def __init__(self, f, g,
                 internals=None,
                 anharmonic_scalings=None,
                 anharmonic_couplings=None,
                 anharmonic_shifts=None,
                 freqs=None,
                 modes=None,
                 dipole_derivatives=None,
                 **operator_settings):
        self.f = np.asanyarray(f)
        self.g = np.asanyarray(g)
        if freqs is None:
            freqs = scipy.linalg.eigvalsh(self.f, self.g, type=3)
            freqs = np.sign(freqs) * np.sqrt(np.abs(freqs))
        self.freqs = np.asanyarray(freqs)
        self.local_freqs = np.diag(f)
        self.basis = HarmonicOscillatorProductBasis(len(f))
        self.modes = modes
        self.rep = 1 / 2 * (
                self.basis.representation(
                    "p", "p",
                    coeffs=self.g,
                    name="T",
                    **operator_settings
                ) +
                self.basis.representation(
                    "x", "x",
                    coeffs=self.f,
                    axes=[[0, 1], [0, 1]],
                    name="V",
                    **operator_settings
                )
        )
        # self.dip_reps: list[ExpansionRepresentation] | None
        if dipole_derivatives is not None:
            dipole_derivatives = [
                self.basis.representation(
                    "x",
                    coeffs=d,
                    axes=[[0], [0]],
                    name="M",
                    **operator_settings
                )
                for d in dipole_derivatives
            ]
        self.dip_reps = dipole_derivatives
        self.internals = internals
        self.scalings = anharmonic_scalings
        self.couplings = anharmonic_couplings
        self.shifts = anharmonic_shifts

    def get_hamiltonian(self, states, coupled_space=None, remove_zpe=True, **rep_opts):
        states = self.prep_states(states)
        if coupled_space is None:
            if isinstance(states, BraKetSpace):
                coupled_space = states
                states = states.bras.take_unique()
            else:
                coupled_space = self.rep.get_transformed_space(states, filter_space=states)[0]
        ham = self.rep.get_representation_matrix(
            coupled_space,
            states,
            zero_element_warning=False,
            **rep_opts
        ).asarray()

        if remove_zpe:
            zpe = np.sum(np.abs(self.freqs)) * 1/2
            ham = ham - np.eye(len(ham)) * zpe

        if (
                self.scalings is not None
                or self.couplings is not None
                or self.shifts is not None
        ):
            if self.internals is not None:
                ham = util.modify_internal_hamiltonian(
                    ham,
                    self.internals,
                    states=states.excitations,
                    scaling_types=self.scalings,
                    coupling_types=self.couplings,
                    shift_types=self.shifts,
                )
            else:
                ham = util.modify_hamiltonian(
                    ham,
                    [tuple(s) for s in states.excitations],
                    scaling_types=self.scalings,
                    coupling_types=self.couplings,
                    shift_types=self.shifts
                )

        return ham

    def get_dipoles(self, states, coupled_space=None, **rep_opts):
        if self.dip_reps is None:
            raise ValueError("`dipole_derivatives` must be supplied to compute dipole matrices")
        states = self.prep_states(states)
        if coupled_space is None:
            if isinstance(states, BraKetSpace):
                coupled_space = states
                states = states.bras.take_unique()
            else:
                coupled_space = self.dip_reps[0].get_transformed_space(states, filter_space=states)[0]
        return np.moveaxis(
            np.array([
                r.get_representation_matrix(
                    coupled_space,
                    states,
                    zero_element_warning=False,
                    **rep_opts
                ).asarray()
                for r in self.dip_reps
            ]),
            0, -1
        )

    def prep_states(self, states):
        if not isinstance(states, AbstractStateSpace):
            if isinstance(states, dict):
                states = states.copy()
                thresh = states.pop("max_freq")
                states = BasisStateSpace.states_under_freq_threshold(
                    self.local_freqs,
                    thresh,
                    **states
                )
                if not isinstance(states, AbstractStateSpace):
                    states = BasisStateSpace(self.basis, states)
                    states = states.as_sorted()
            elif nput.is_int(states):
                states = BasisStateSpace.from_quanta(self.basis, states)
            else:
                if (
                        nput.is_numeric_array_like(states)
                    and np.asanyarray(states).ndim == 2
                ):
                    states = BasisStateSpace(self.basis, states)
                else:
                    s0 = self.prep_states(states[0])
                    for s1 in states[1:]:
                        s0 = s0.union(self.prep_states(s1))
                    states = s0

        return states
    def get_wavefunctions(self, states, coupled_space=None, remove_zpe=True, include_dipoles=True,
                          expansion_opts=None,
                          **wfn_opts
                          ):
        if expansion_opts is None:
            expansion_opts = {}

        states = self.prep_states(states)

        ham = self.get_hamiltonian(states, coupled_space=coupled_space, remove_zpe=remove_zpe, **expansion_opts)
        engs, coeffs = np.linalg.eigh(ham)
        ground_state = BasisStateSpace.from_quanta(self.basis, 0)
        gs_loc = states.find(ground_state, missing_val=-1)[0]
        if gs_loc == -1:
            if remove_zpe:
                zpe = 0
            else:
                zpe = np.sum(np.abs(self.freqs)) * 1/2
            engs = np.concatenate([[zpe], engs])
            coeffs = np.pad(coeffs, [[1, 0], [1, 0]])
            coeffs[0, 0] = 1
            states = ground_state.union(states)
        elif gs_loc > 0:
            engs = np.concatenate([engs[(gs_loc,)], engs[:gs_loc], engs[gs_loc+1:]])
            coeffs = np.concatenate([coeffs[:, (gs_loc,)], engs[:, :gs_loc], engs[:, gs_loc+1:]], axis=1)


        if include_dipoles:
            dips = self.get_dipoles(states, coupled_space=coupled_space, **expansion_opts)
        else:
            dips = None

        return LocalHarmonicWavefunctions(self,
                                          engs, coeffs,
                                          basis=states,
                                          dipole_matrix=dips,
                                          hamiltonian=ham,
                                          **wfn_opts
                                          )

    @classmethod
    def _match_mode_label(cls, test, freq, label):
        if label is None: return test is None
        if len(test) == 2 and (
                not (test[0] is None or isinstance(test[0], str))
        ):
            if nput.is_numeric(test[0]):
                test_freqs = test
                test = None
            else:
                test_freqs, test = test
        else:
            test_freqs = None

        if test_freqs is not None:
            lf, mF = test_freqs
            if freq < lf or freq > mF: return False

        if test is None: return True

        if isinstance(test, str):
            test = [test]
        n_test = len(test)
        if len(label) < n_test: return False
        return all(
            t is None or (t == l if isinstance(t, str) else l in t)
            for t,l in zip(test, label[-n_test:])
        )

    @classmethod
    def localize_internal_modes(cls,
                                nms, internals,
                                localization_mode_spaces=None,
                                mode_labels=None,
                                use_nonnegative_modes=True
                                ):

        if use_nonnegative_modes:
            neg_modes = np.where(nms.freqs < 1e-4 / UnitsData.hartrees_to_wavenumbers)
            if len(neg_modes) > 0 and len(neg_modes[0]) > 0:
                neg_modes = neg_modes[0]
                rem = np.setdiff1d(np.arange(len(nms.freqs)), neg_modes)
                nms = nms[rem]
                if mode_labels is not None:
                    if hasattr(mode_labels, 'items'):
                        mode_labels = dev.dict_take(mode_labels, rem)
                    else:
                        mode_labels = [mode_labels[r] for r in rem]

        if isinstance(internals, dict):

            if mode_labels is not None:
                # for i, (freq, lab) in enumerate(zip(reversed(nms.freqs), reversed(mode_labels))):
                #     print(
                #         "Mode {} ({}): {:.0f} {}".format(i + 1,
                #                                          len(nms.freqs) - (i+1),
                #                                          freq * UnitsData.hartrees_to_wavenumbers,
                #                                          "mixed"
                #                                          if lab.type is None else
                #                                          lab.type
                #                                          )
                #     )

                internals = internals.copy()
                mod_internals = {}
                for coord, c in internals.items():
                    if (
                            len(c) == 2
                            and not (c[0] is None or isinstance(c[0], str))
                    ):
                        mod_internals[coord] = (
                            c[0],
                            [
                                i for i,(f,l) in enumerate(zip(nms.freqs, mode_labels))
                                if cls._match_mode_label(c[1], f, l)
                            ]
                        )
                internals.update(mod_internals)

                if localization_mode_spaces is not None:
                    localization_mode_spaces = {
                        k: (
                            lm
                                if all(nput.is_int(i) for i in lm) else
                            [
                                i
                                    for i, (f,l) in enumerate(zip(nms.freqs, mode_labels))
                                if cls._match_mode_label(lm, f, l.type if l is not None else l)
                            ]
                        )
                        for k,lm in localization_mode_spaces.items()
                    }

            dim = len(nms.freqs)
            spaces = {}
            subints = {}
            subspaces_used = False
            for n, (coord, c) in enumerate(internals.items()):
                if (
                        len(c) == 2
                        and all(nput.is_int(i) for i in c[1])
                ):
                    subspaces_used = True
                    key = tuple(sorted(c[1]))
                    if key not in spaces:
                        spaces[key] = {}
                    spaces[key][coord] = n
                    subints[coord] = c[0]
                elif localization_mode_spaces is not None:
                    space = localization_mode_spaces.get(coord)
                    if space is not None:
                        subspaces_used = True
                        key = tuple(space)
                        if key not in spaces:
                            spaces[key] = {}
                        spaces[key][coord] = n
                        subints[coord] = c
                    elif isinstance(c, str):
                        space = localization_mode_spaces.get(c)
                        if space is not None:
                            subspaces_used = True
                            key = tuple(space)
                            if key not in spaces:
                                spaces[key] = {}
                            spaces[key][coord] = n
                            subints[coord] = c
                        else:
                            key = tuple(range(dim))
                            if key not in spaces:
                                spaces[key] = {}
                            spaces[key][coord] = n
                            subints[coord] = c
                    else:
                        for i in range(len(c)):
                            sub = tuple(c[i:])
                            space = localization_mode_spaces.get(sub)
                            if space is not None:
                                subspaces_used = True
                                key = tuple(space)
                                if key not in spaces:
                                    spaces[key] = {}
                                spaces[key][coord] = n
                                subints[coord] = c
                                break
                        else:
                            key = tuple(range(dim))
                            if key not in spaces:
                                spaces[key] = {}
                            spaces[key][coord] = n
                            subints[coord] = c

                else:
                    key = tuple(range(dim))
                    if key not in spaces:
                        spaces[key] = {}
                    spaces[key][coord] = n
                    subints[coord] = c

            # noneg_selection = None
            # if use_nonnegative_modes:
            #     noneg_selection = np.where(nms.freqs < 1e-4 / UnitsData.hartrees_to_wavenumbers)
            #     if len(noneg_selection) > 0 and len(noneg_selection[0]) > 0:
            #         subspaces_used = True
            #         noneg_selection= noneg_selection[0]


            internals = subints
            if subspaces_used:
                ord = []
                tranfs = []
                for space, ints in spaces.items():
                    space = np.array([
                        dim + s if s < 0 else s
                        for s in space
                    ])
                    # if noneg_selection is not None:
                    #     space = np.setdiff1d(space, noneg_selection)

                    subnms = nms[space]
                    sublocs = subnms.localize(internals=ints)
                    tf = np.zeros((dim, len(ints)))
                    inv = np.zeros((len(ints), dim))
                    t_sub, inv_sub = sublocs.localizing_transformation
                    tf[space, :] = t_sub
                    inv[:, space] = inv_sub
                    tranfs.append((tf, inv))
                    ord.extend(ints.values())

                tf = np.concatenate([t for t, i in tranfs], axis=1)
                inv = np.concatenate([i for t, i in tranfs], axis=0)
                ord = np.argsort(ord)
                tf = tf[:, ord]
                inv = inv[ord, :]
                loc_modes = nms.apply_transformation((tf, inv))
            else:
                loc_modes = nms.localize(internals=internals)
        else:
            # noneg_selection = None
            # if use_nonnegative_modes:
            #     noneg_selection = np.where(nms.freqs < 1e-4 / UnitsData.hartrees_to_wavenumbers)
            #     if len(noneg_selection) > 0 and len(noneg_selection[0]) > 0:
            #         noneg_selection = noneg_selection[0]
            #     else:
            #         noneg_selection = None
            #
            # if noneg_selection is not None:
            #     dim = len(nms.freqs)
            #     subnms = nms[noneg_selection]
            #     sublocs = subnms.localize(internals=internals)
            #     tf = np.zeros((dim, len(internals)))
            #     inv = np.zeros((len(internals), dim))
            #     t_sub, inv_sub = sublocs.localizing_transformation
            #     tf[noneg_selection, :] = t_sub
            #     inv[:, noneg_selection] = inv_sub
            #     loc_modes = nms.apply_transformation((tf, inv))
            # else:
            loc_modes = nms.localize(internals=internals)

        return loc_modes

    @classmethod
    def from_modes(cls,
                   nms,
                   internals=None,
                   mode_labels=None,
                   localization_mode_spaces=None,
                   oblique=True,
                   include_complement=False,
                   dipole_derivatives=None,
                   localize=True,
                   use_nonnegative_modes=True,
                   **opts):
        from ..Modes import NormalModes
        nms:NormalModes

        nms = nms.remove_frequency_scaling().remove_mass_weighting()
        if localize:
            loc_modes = cls.localize_internal_modes(
                nms,
                internals,
                localization_mode_spaces=localization_mode_spaces,
                mode_labels=mode_labels,
                use_nonnegative_modes=use_nonnegative_modes
            )

            if oblique:
                loc_modes = loc_modes.make_oblique()
            if include_complement:
                loc_modes = loc_modes.get_complement(concatenate=True)
                if internals is not None:
                    k = len(internals)
                    M = len(loc_modes.freqs)
                    if k < M:
                        if isinstance(internals, dict):
                            internals = internals.copy()
                            for i in range(k, M):
                                internals[i] = coordops.mode_label(
                                    None,
                                    None,
                                    None,
                                    "complement"
                                )
                        else:
                            internals = list(internals) + list(range(k, M))
        else:
            loc_modes = nms

        if dipole_derivatives is not None:
            # freq_conv = np.sqrt(loc_modes.compute_freqs())
            L = loc_modes.coords_by_modes
            if not oblique:
                a = np.diag(1 / np.sqrt(loc_modes.local_freqs))
                L = a @ L
            dipole_derivatives = [ L @ d for d in dipole_derivatives]


        # loc_modes = loc_modes.make_mass_weighted()
        f = loc_modes.local_hessian
        g = loc_modes.local_gmatrix


        return cls(
            f, g,
            internals=internals,
            dipole_derivatives=dipole_derivatives,
            modes=loc_modes,
            **opts
        )

    @classmethod
    def state(cls, nquanta_or_state, *tags, indices=None):
        if isinstance(nquanta_or_state, util.product_state):
            return nquanta_or_state
        nquanta = nquanta_or_state
        if nput.is_int(nquanta):
            if all(nput.is_int(q) for q in tags):
                return (nquanta,) + tags
        elif isinstance(nquanta, str):
            tags = (nquanta,) + tags
            nquanta = None
        elif nquanta is not None:
            if (
                    all(q is None or isinstance(q, str) for q in nquanta)
                    and len(tags) == 0
            ):
                return cls.state(None, *nquanta, indices=indices)
            elif nput.is_int(nquanta[0]):
                return cls.state(*nquanta, *tags, indices=indices)
            elif len(tags) == 0:
                base_data = [cls.state(n) for n in nquanta]
                indices = tuple(b.indices[0] if b.indices is not None else None for b in base_data)
                nquanta = tuple(b.quanta[0] if b.quanta is not None else None for b in base_data)
                tags = tuple(b.tags[0] if b.tags is not None else None for b in base_data)
            else:
                base_data = [cls.state(nquanta)] + [cls.state(t) for t in tags]
                indices = tuple(b.indices[0] if b.indices is not None else None for b in base_data)
                nquanta = tuple(b.quanta[0] if b.quanta is not None else None for b in base_data)
                tags = tuple(b.tags[0] if b.tags is not None else None for b in base_data)

        if isinstance(tags, str):
            tags = ((tags,),)
        elif all(q is None or isinstance(q, str) for q in tags):
            tags = (tags,)
        if indices is not None:
            if nput.is_int(indices):
                indices = (indices,)
            elif all(i is None for i in indices):
                indices = None
        if nquanta is not None:
            if nput.is_int(nquanta):
                nquanta = (nquanta,)
            elif all(i is None for i in nquanta):
                nquanta = None
        return util.product_state(indices, nquanta, tags)

    @classmethod
    def state_pair(cls, shared_atoms, state_1, state_2=None):
        if not (
                shared_atoms is None
                or nput.is_int(shared_atoms)
                or all(nput.is_int(a) for a in shared_atoms)
        ) and state_2 is None:
            state_1, state_2 = shared_atoms, state_1
            shared_atoms = None
        if shared_atoms is not None:
            return (
                shared_atoms,
                cls.state(state_1),
                cls.state(state_2),
            )
        else:
            return (
                cls.state(state_1),
                cls.state(state_2),
            )

    @classmethod
    def from_molecule(cls,
                      mol,
                      modes=None,
                      mode_labels=None,
                      internals=None,
                      coordinate_filter=None,
                      allowed_coordinate_types=None,
                      excluded_coordinate_types=None,
                      allowed_ring_types=None,
                      excluded_ring_types=None,
                      allowed_group_types=None,
                      excluded_group_types=None,
                      include_stretches=True,
                      include_bends=True,
                      include_dihedrals=False,
                      dipole_derivatives=None,
                      include_dipole=True,
                      internal_coordinate_sorting=None,
                      prune_excess_internals=False,
                      **opts):
        from ..Molecools import Molecule
        mol: Molecule
        if modes is None:
            modes = mol.get_normal_modes(use_internals=False)


        if internals is None:
            internals = mol.get_labeled_internals(
                coordinate_filter=coordinate_filter,
                allowed_coordinate_types=allowed_coordinate_types,
                excluded_coordinate_types=excluded_coordinate_types,
                allowed_ring_types=allowed_ring_types,
                excluded_ring_types=excluded_ring_types,
                allowed_group_types=allowed_group_types,
                excluded_group_types=excluded_group_types,
                include_stretches=include_stretches,
                include_bends=include_bends,
                include_dihedrals=include_dihedrals,
                coordinate_sorting=internal_coordinate_sorting,
                pruning=prune_excess_internals
            )

        if mode_labels is True:
            mode_labels = mol.get_mode_labels(
                # internals,
                modes=modes,
                pruning=prune_excess_internals,
                use_redundants=not prune_excess_internals
            )

        if dipole_derivatives is None and include_dipole:
            if mol.dipole_derivatives is not None:
                dipole_derivatives = mol.dipole_derivatives[1].T

        return cls.from_modes(
            modes,
            internals=internals,
            mode_labels=mode_labels,
            dipole_derivatives=dipole_derivatives,
            **opts
        )

class LocalHarmonicWavefunctions(MatrixWavefunctions):
    def __init__(self, model, energies, coefficients, **etc):
        super().__init__(energies, coefficients, **etc)
        self.model = model

    def get_modification_dict(self,
                              *,
                              model=None,
                              energies=None,
                              wavefunctions=None,
                              **etc
                              ):
        return dict(
            dict(model=model if model is not None else self.model),
            **super().get_modification_dict(
                energies=energies,
                wavefunctions=wavefunctions,
                **etc
            )
        )

    def modify(self,
               *,
               model=None,
               energies=None,
               wavefunctions=None,
               **etc
               ):
        mod_dict = self.get_modification_dict(
            model=model,
            energies=energies,
            wavefunctions=wavefunctions,
            **etc
        )
        return type(self)(
            mod_dict.pop('model'),
            mod_dict.pop('energies'),
            mod_dict.pop('wavefunctions'),
            **mod_dict
        )

class CustomLocalHarmonicModel(LocalHarmonicModel):
    default_localization_settings = {}
    default_molecule_settings = {}
    default_scalings = {}
    default_couplings = {}
    default_shifts = {}

    def __init__(self,
                 f, g,
                 internals=None,
                 anharmonic_scalings=None,
                 anharmonic_couplings=None,
                 anharmonic_shifts=None,
                 dipole_derivatives=None,
                 **etc
                 ):
        if anharmonic_scalings is None:
            anharmonic_scalings = {}
        anharmonic_scalings = dev.merge_dicts(self.default_scalings, anharmonic_scalings)
        if anharmonic_couplings is None:
            anharmonic_couplings = {}
        anharmonic_couplings = dev.merge_dicts(self.default_couplings, anharmonic_couplings)
        if anharmonic_shifts is None:
            anharmonic_shifts = {}
        anharmonic_shifts = dev.merge_dicts(self.default_shifts, anharmonic_shifts)

        super().__init__(
            f, g,
            internals=internals,
            anharmonic_scalings=anharmonic_scalings,
            anharmonic_couplings=anharmonic_couplings,
            anharmonic_shifts=anharmonic_shifts,
            dipole_derivatives=dipole_derivatives,
            **etc
        )

    @classmethod
    def from_modes(cls,
                   nms,
                   internals=None,
                   dipole_derivatives=None,
                   **opts):
        opts = dev.merge_dicts(cls.default_localization_settings, opts)
        return super().from_modes(
            nms,
            internals=internals,
            dipole_derivatives=dipole_derivatives,
            **opts
        )

    @classmethod
    def from_molecule(cls,
                      mol,
                      modes=None,
                      internals=None,
                      dipole_derivatives=None,
                      **opts):
        opts = dev.merge_dicts(cls.default_molecule_settings, opts)
        return super().from_molecule(
            mol,
            modes=modes,
            internals=internals,
            dipole_derivatives=dipole_derivatives,
            **opts
        )


class TaborCHModel(CustomLocalHarmonicModel):
    default_localization_settings = {"oblique":False}
    default_molecule_settings = dict(
        allowed_coordinate_types={'CH', 'HCH', 'NH', 'OH'},
        excluded_ring_types={'benzene'}
    )
    default_scalings = {
        LocalHarmonicModel.state("NH", "stretch"): 0.961,
        LocalHarmonicModel.state("methyl", "CH", "stretch"): 0.961,
        LocalHarmonicModel.state("ethyl", "CH", "stretch"): 0.961,
        LocalHarmonicModel.state("HCH", "bend"): 0.975,  # fundamentals and overtones
        LocalHarmonicModel.state(("HCH", "bend"), ("HCH", "bend")): 0.975,  # combination band
    }
    default_couplings = {
        # CH to adjacent bend overtone
        LocalHarmonicModel.state_pair(
            2,  # shared atoms
            ("CH", "stretch"),  # stretch fundamental
            (2, ("HCH", "bend"))  # bend overtone
        ): 22 / UnitsData.hartrees_to_wavenumbers,

        # CH3 to opposite bend combination
        LocalHarmonicModel.state_pair(
            ((2, 1),),  # shared atoms
            ("CH", "stretch"),  # stretch fundamental
            (("HCH", "bend"), ("HCH", "bend"))  # bend overtone
        ): 5.6 / UnitsData.hartrees_to_wavenumbers,
        LocalHarmonicModel.state_pair(
            ((1, 2),),  # shared atoms
            ("CH", "stretch"),  # stretch fundamental
            (("HCH", "bend"), ("HCH", "bend"))  # bend overtone
        ): 5.6 / UnitsData.hartrees_to_wavenumbers,

        # CH3 to adjacent bend combination
        LocalHarmonicModel.state_pair(
            ((2, 2),),  # shared atoms
            ("CH", "stretch"),  # stretch fundamental
            (("HCH", "bend"), ("HCH", "bend"))  # bend overtone
        ): 1.5 / UnitsData.hartrees_to_wavenumbers,
        LocalHarmonicModel.state_pair(
            ((2, 2),),  # shared atoms
            ("CH", "stretch"),  # stretch fundamental
            (("HCH", "bend"), ("HCH", "bend"))  # bend overtone
        ): 1.5 / UnitsData.hartrees_to_wavenumbers
    }
    default_shifts = {
        # internal-matching localization overshoots mildly
        ("ethyl", "CH", "stretch"): -4 / UnitsData.hartrees_to_wavenumbers,
        # internal-matching localization requires less bend overtone shifting
        ("HCH", "bend"): -2 * 9.6 / UnitsData.hartrees_to_wavenumbers
    }


import numpy as np

import McUtils.Plots as plt
import McUtils.Numputils as nput

__all__ = [
    "plot_energy_levels"
]

def plot_energy_levels(energy_list, figure: plt.Graphics = None,
                       x_list=None,
                       bar_spacing=.025,
                       bar_width=None,
                       plot_range=None,
                       color='black',
                       ticks=None,
                       labels=None,
                       bar_styles=None,
                       **styles):

    if isinstance(energy_list, dict): # replace with collections.Mapping...
        labels = energy_list.keys()
        energy_list = list(energy_list.values())

    if isinstance(energy_list[0], dict):
        if 'x' in energy_list[0]:
            x_list = [e['x'] for e in energy_list]
        if bar_styles is None:
            bar_styles = [None] * len(energy_list)
        _ = []
        for el, b in zip(energy_list, bar_styles):
            if b is None: b = {}
            b = dict(el, **b)
            b.pop('x', None)
            b.pop('y')
            _.append(b)
        bar_styles = _
        energy_list = [e['y'] for e in energy_list]

    if plot_range is None and figure is not None:
        plot_range = figure.plot_range
    if plot_range is not None and not nput.is_numeric(plot_range[0]):
        x_range = plot_range[0]
    else:
        if x_list is not None:
            x_list = np.asanyarray(x_list)
            x_min = np.min(x_list)
            x_max = np.max(x_list)
            diff = (x_max - x_min) / (1 - 2*bar_spacing)
            x_range = [x_min - bar_spacing * diff, x_max + bar_spacing * diff]
        else:
            x_range = [0, 1]
    if x_list is None:
        neng = len(energy_list)
        x_list = [[i / neng + bar_spacing, (i + 1) / neng - bar_spacing] for i in range(neng)]
    elif nput.is_numeric(x_list[0]):
        x_list = np.asanyarray(x_list)
        x_min = np.min(x_list)
        x_max = np.max(x_list)
        if bar_width is None:
            bar_width = np.min(np.diff(np.sort(x_list)))
        pad = bar_width / 2
        x_range = [x_min - pad, x_max + pad]
        s = bar_spacing * (x_range[-1] - x_range[0])
        x_list = [[x - pad + s, x + pad - s] for x in x_list]

    x_list = np.asanyarray(x_list)
    x_min = np.min(x_list)
    x_max = np.max(x_list)
    diff = (x_max - x_min) / (1 - 2 * bar_spacing)
    cur_range = [x_min - bar_spacing * diff, x_max + bar_spacing * diff]
    x_list = np.reshape(nput.vec_rescale(x_list.flatten(), x_range, cur_range=cur_range), x_list.shape)

    if plot_range is None or nput.is_numeric(plot_range[0]):
        plot_range = [x_range, plot_range]

    if ticks is None or nput.is_numeric(ticks[0]):
        if labels is None:
            x_ticks = []
        else:
            tick_pos = [(a+b)/2 for a,b in x_list]
            x_ticks = (tick_pos, dict(labels=labels))
        ticks = [x_ticks, ticks]
    elif labels is not None:
        x_ticks, y_ticks = ticks
        if nput.is_numeric(x_ticks[0]):
            x_ticks = (x_ticks, dict(labels=labels))
        ticks = [x_ticks, y_ticks]

    if bar_styles is None:
        bar_styles = {}
    if isinstance(bar_styles, dict):
        bar_styles = [bar_styles] * len(energy_list)
    if color is None or isinstance(color, str):
        color = [color] * len(energy_list)

    for i, (f, x, c, s) in enumerate(zip(energy_list, x_list, color, bar_styles)):
        style_dict = dict(styles, **s)
        c = style_dict.pop('color', c)
        figure = plt.HorizontalLinePlot(
            x,
            f,
            figure=figure,
            color=c,
            plot_range=plot_range,
            ticks=ticks,
            **style_dict
        )
    return figure